import anndata
import os
import pandas as pd

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE121nnn/GSE121862/suppl/" \
            "GSE121862%5FUCSD%2DWU%5FSingle%5FNuclei%5FCluster%5FAnnotated%5FRaw%5FUMI%5FMatrix%2Etsv%2Egz"
        self.download_url_meta = "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE121nnn/GSE121862/suppl/" \
            "GSE121862%5FUCSD%2DWU%5FSingle%5FNuclei%5FCluster%5FAnnotations%2Ecsv%2Egz"

        self.assay_sc = "DroNc-seq"
        self.author = "Lake"
        self.disease = "healthy"
        self.doi_journal = "10.1038/s41467-019-10861-2"
        self.layer_counts = "X"
        self.organ = "kidney"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.year = 2019

        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"
        self.cell_type_obs_key = "celltype"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = [
        os.path.join(data_dir, "GSE121862_UCSD-WU_Single_Nuclei_Cluster_Annotated_Raw_UMI_Matrix.tsv.gz"),
        os.path.join(data_dir, "GSE121862_UCSD-WU_Single_Nuclei_Cluster_Annotations.csv.gz")
    ]
    adata = anndata.AnnData(pd.read_csv(fn[0], sep="\t").T)
    annot = pd.read_csv(fn[1], index_col=0, dtype="category")
    adata.obs["celltype"] = [annot.loc[i.split("_")[0][1:]]["Annotation"] for i in adata.obs.index]

    return adata



import anndata
import gzip
import os
import shutil
from tempfile import TemporaryDirectory


def load(data_dir, sample_fn, **kwargs):
    fn = os.path.join(data_dir, sample_fn)
    with TemporaryDirectory() as tmpdir:
        tmppth = tmpdir + "/decompressed.h5ad"
        with gzip.open(fn, "rb") as input_f, open(tmppth, "wb") as output_f:
            shutil.copyfileobj(input_f, output_f)
        adata = anndata.read_h5ad(tmppth)
    adata.var["feature_types"] = [
        {"ATAC": "peak", "GEX": "rna", "ADT": "protein"}[x]
        for x in adata.var["feature_types"].values
    ]
    return adata



import anndata
import os
import pandas as pd
import scipy.io
import gzip
import tarfile

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE131nnn/GSE131685/suppl/GSE131685_RAW.tar"
        self.download_url_meta = None

        self.assay_sc = "10x 3' v2"
        self.author = "Liao"
        self.disease = "healthy"
        self.layer_counts = "X"
        self.organ = "kidney"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.year = 2020
        self.doi_journal = "10.1038/s41597-019-0351-8"

        self.feature_id_var_key = "ensembl"
        self.feature_symbol_var_key = "names"
        self.feature_type = "rna"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = os.path.join(data_dir, "GSE131685_RAW.tar")
    adatas = []
    with tarfile.open(fn) as tar:
        for member in tar.getmembers():
            if "_matrix.mtx.gz" in member.name:
                name = "_".join(member.name.split("_")[:-1])
                with gzip.open(tar.extractfile(member), "rb") as mm:
                    X = scipy.io.mmread(mm).T.tocsr()
                obs = pd.read_csv(tar.extractfile(name + "_barcodes.tsv.gz"), compression="gzip", header=None,
                                  sep="\t", index_col=0)
                obs.index.name = None
                var = pd.read_csv(tar.extractfile(name + "_features.tsv.gz"), compression="gzip", header=None,
                                  sep="\t").iloc[:, :2]
                var.columns = ["ensembl", "names"]
                var.index = var["ensembl"].values
                adata = anndata.AnnData(X=X, obs=obs, var=var)
                adata.obs["sample"] = name
                adatas.append(adata)
    adata = adatas[0].concatenate(adatas[1:])

    return adata



import numpy as np
try:
    import tensorflow as tf
except ImportError:
    tf = None
from typing import List, Union, Tuple

from sfaira.models.embedding.output_layers import NegBinOutput, NegBinSharedDispOutput, NegBinConstDispOutput, \
    GaussianOutput, GaussianSharedStdOutput, GaussianConstStdOutput
from sfaira.versions.topologies import TopologyContainer
from sfaira.models.embedding.base import BasicModelKerasEmbedding
from sfaira.models.pp_layer import PreprocInput


class Sampling(tf.keras.layers.Layer):
    """Uses (z_mean, z_log_var) to sample z."""

    def call(self, inputs, **kwargs):
        z_mean, z_log_var = inputs
        batch = tf.shape(z_mean)[0]
        dim = tf.shape(z_mean)[1]
        epsilon = tf.keras.backend.random_normal(shape=(batch, dim))
        return z_mean + tf.exp(0.5 * z_log_var) * epsilon


class Encoder(tf.keras.layers.Layer):
    """Maps input to embedding space"""

    def __init__(
            self,
            latent_dim: Tuple,
            dropout_rate,
            l1_coef: float,
            l2_coef: float,
            batchnorm: bool,
            activation: str,
            kernel_initializer: str,
            name='encoder',
            **kwargs
    ):
        super().__init__(name=name, **kwargs)

        self.fwd_pass = []
        for i, hid_size in enumerate(latent_dim[:-1]):
            self.fwd_pass.append(
                tf.keras.layers.Dense(
                    units=hid_size,
                    activation=activation,
                    kernel_initializer=kernel_initializer,
                    kernel_regularizer=tf.keras.regularizers.l1_l2(l1=l1_coef, l2=l2_coef),
                    name='enc%s' % i
                )
            )
            # At the bottleneck layer:
            # 1. a linear activation function is used to not restrict the support of the hidden units
            # 2. no batch normalisation is used to not scale the variance of inactive units up and the activity of
            # active units down.
            # 3. no dropout is used to not confound the required bottleneck dimension with dropout rate.
            if batchnorm and i < (len(latent_dim) - 1):
                self.fwd_pass.append(
                    tf.keras.layers.BatchNormalization(
                        center=True,
                        scale=True
                    )
                )
            if i < (len(latent_dim) - 1) and dropout_rate > 0:
                if activation == "selu":
                    self.fwd_pass.append(tf.keras.layers.AlphaDropout(dropout_rate))
                else:
                    self.fwd_pass.append(tf.keras.layers.Dropout(dropout_rate, noise_shape=None, seed=None))

        # final layer
        self.dense_mean = tf.keras.layers.Dense(
            units=latent_dim[-1],
            activation="linear"
        )
        self.dense_log_var = tf.keras.layers.Dense(
            units=latent_dim[-1],
            activation="linear"
        )
        self.sampling = Sampling()

    def call(self, inputs, **kwargs):
        x = inputs
        for layer in self.fwd_pass:
            x = layer(x)
        # final layer
        z_mean = self.dense_mean(x)
        z_log_var = self.dense_log_var(x)
        z = self.sampling((z_mean, z_log_var))
        return z, z_mean, z_log_var


class Decoder(tf.keras.layers.Layer):
    """Maps latent space sample back to output"""

    def __init__(
            self,
            latent_dim: Tuple,
            dropout_rate,
            l1_coef: float,
            l2_coef: float,
            batchnorm: bool,
            activation: str,
            kernel_initializer: str,
            name='decoder',
            **kwargs
    ):
        super().__init__(name=name, **kwargs)

        self.fwd_pass = []
        for i, hid_size in enumerate(latent_dim):
            self.fwd_pass.append(
                tf.keras.layers.Dense(
                    units=hid_size,
                    activation=activation,
                    kernel_initializer=kernel_initializer,
                    kernel_regularizer=tf.keras.regularizers.l1_l2(l1=l1_coef, l2=l2_coef),
                    name='dec%s' % i
                )
            )
            if batchnorm:
                self.fwd_pass.append(
                    tf.keras.layers.BatchNormalization(
                        center=True,
                        scale=True
                    )
                )

            if dropout_rate > 0.0:
                if activation == "selu":
                    self.fwd_pass.append(tf.keras.layers.AlphaDropout(dropout_rate))
                else:
                    self.fwd_pass.append(tf.keras.layers.Dropout(dropout_rate, noise_shape=None, seed=None))

    def call(self, inputs, **kwargs):
        x = inputs
        for layer in self.fwd_pass:
            x = layer(x)
        return x


class ModelKerasVae(BasicModelKerasEmbedding):

    def predict_reconstructed(self, x: np.ndarray):
        return np.split(self.training_model.predict(x)[0], indices_or_sections=2, axis=1)[0]

    def __init__(
            self,
            in_dim,
            latent_dim=(128, 64, 2, 64, 128),
            dropout_rate=0.1,
            l1_coef: float = 0.,
            l2_coef: float = 0.,
            batchnorm: bool = False,
            activation='tanh',
            init='glorot_uniform',
            output_layer="nb"
    ):
        super(ModelKerasVae, self).__init__()
        # Check length of latent dim to divide encoder-decoder stack:
        if len(latent_dim) % 2 == 1:
            n_layers_enc = len(latent_dim) // 2 + 1
        else:
            raise ValueError("len(latent_dim)=%i should be uneven to provide a defined bottleneck" % len(latent_dim))

        inputs_encoder = tf.keras.Input(shape=(in_dim,), name='counts')
        inputs_sf = tf.keras.Input(shape=(1,), name='size_factors')
        inputs_encoder_pp = PreprocInput()(inputs_encoder)
        output_encoder = Encoder(
            latent_dim=latent_dim[:n_layers_enc],
            dropout_rate=dropout_rate,
            l1_coef=l1_coef,
            l2_coef=l2_coef,
            batchnorm=batchnorm,
            activation=activation,
            kernel_initializer=init
        )(inputs_encoder_pp)

        z, z_mean, z_log_var = output_encoder
        expected_logqz_x = -0.5 * tf.reduce_sum(1 + z_log_var, axis=1)
        expected_logpz = -0.5 * tf.reduce_sum(tf.square(z_mean) + tf.exp(z_log_var), axis=1)

        expected_densities = tf.keras.layers.Concatenate(axis=1, name='kl')([
            tf.expand_dims(expected_logqz_x, axis=1),
            tf.expand_dims(expected_logpz, axis=1)])

        output_decoder = Decoder(
            latent_dim=latent_dim[n_layers_enc:],
            dropout_rate=dropout_rate,
            l1_coef=l1_coef,
            l2_coef=l2_coef,
            batchnorm=batchnorm,
            activation=activation,
            kernel_initializer=init
        )(z)

        if output_layer == 'nb':
            output_decoder_expfamily = NegBinOutput(original_dim=in_dim)((output_decoder, inputs_sf))
        elif output_layer == 'nb_shared_disp':
            output_decoder_expfamily = NegBinSharedDispOutput(original_dim=in_dim)((output_decoder, inputs_sf))
        elif output_layer == 'nb_const_disp':
            output_decoder_expfamily = NegBinConstDispOutput(original_dim=in_dim)((output_decoder, inputs_sf))
        elif output_layer == 'gaussian':
            output_decoder_expfamily = GaussianOutput(original_dim=in_dim)((output_decoder, inputs_sf))
        elif output_layer == 'gaussian_shared_disp':
            output_decoder_expfamily = GaussianSharedStdOutput(original_dim=in_dim)((output_decoder, inputs_sf))
        elif output_layer == 'gaussian_const_disp':
            output_decoder_expfamily = GaussianConstStdOutput(original_dim=in_dim)((output_decoder, inputs_sf))
        else:
            raise ValueError("tried to access a non-supported output layer %s" % output_layer)
        output_decoder_expfamily_concat = tf.keras.layers.Concatenate(axis=1, name="neg_ll")(output_decoder_expfamily)

        self.encoder_model = tf.keras.Model(
            inputs=[inputs_encoder, inputs_sf],
            outputs=[z, z_mean, z_log_var],
            name="encoder_model"
        )
        self.training_model = tf.keras.Model(
            inputs=[inputs_encoder, inputs_sf],
            outputs=[output_decoder_expfamily_concat, expected_densities],
            name="autoencoder"
        )

    def predict_embedding(self, x, variational=False):
        if variational:
            return self.encoder_model.predict(x)
        else:
            return self.encoder_model.predict(x)[1]


class ModelVaeVersioned(ModelKerasVae):
    def __init__(
            self,
            topology_container: TopologyContainer,
            override_hyperpar: Union[dict, None] = None
    ):
        hyperpar = topology_container.topology["hyper_parameters"]
        if override_hyperpar is not None:
            for k in list(override_hyperpar.keys()):
                hyperpar[k] = override_hyperpar[k]
        super().__init__(
            in_dim=topology_container.n_var,
            **hyperpar
        )
        print('passed hyperpar: \n', hyperpar)
        self._topology_id = topology_container.topology_id
        self.genome_size = topology_container.n_var
        self.model_class = "embedding"
        self.model_type = topology_container.model_type
        self.hyperparam = dict(
            list(hyperpar.items()) +  # noqa: W504
            [
                ("topology_id", self._topology_id),
                ("genome_size", self.genome_size),
                ("model_class", self.model_class),
                ("model_type", self.model_type)
            ]
        )



from typing import Union, Tuple, Dict

import pandas as pd


def split_batch(x: Union[Tuple, Dict]):
    """
    Splits retrieval batch into consumption batches of length 1.

    Often, end-user consumption batches would be observation-wise, ie yield a first dimension of length 1.

    :param x: Tuple or Dict
        One of the following:
            * Data tuple of length 1 or 2: (input,) or (input, output,), where both input and output are also a tuple,
            but of batch-dimensioned tensors.
            * Dict
    """
    if isinstance(x, tuple):
        batch_dim = x[0][0].shape[0]
        for i in range(batch_dim):
            output = []
            for y in x:
                if isinstance(y, tuple):
                    output.append(tuple([z.iloc[[i], :] if isinstance(z, pd.DataFrame) else z[i, :] for z in y]))
                else:
                    output.append(y.iloc[[i], :] if isinstance(y, pd.DataFrame) else y[i, :])
            yield tuple(output)
    elif isinstance(x, dict):
        keys = list(x.keys())
        batch_dim = x[keys[0]].shape[0]
        for i in range(batch_dim):
            yield {key: x[key].iloc[[i], :] if isinstance(x[key], pd.DataFrame) else x[key][i, :] for key in keys}
    else:
        raise ValueError('Input to split_batch(x) has to be either a Tuple or a Dict')



"""
The classes in this file are containers of field names and element entries that are used in streamlined adata objects
in sfaira and in associated data bases.
"""
from typing import List, Union


class AdataIds:
    """
    Base class of constant field names of anndata.AnnData object entries, such as .uns keys and .obs columns.
    """
    annotated: str
    assay_sc: str
    author: str
    cell_type: str
    development_stage: str
    disease: str
    doi_journal: str
    doi_preprint: str
    download_url_data: str
    download_url_meta: str
    dataset: str
    dataset_group: str
    ethnicity: str
    feature_id: str
    feature_biotype: str
    feature_index: str
    feature_reference: str
    feature_symbol: str
    id: str
    individual: str
    ncells: str
    normalization: str
    organ: str
    organism: str
    primary_data: str
    sample_source: str
    sex: str
    state_exact: str
    tech_sample: str
    year: str

    feature_kwargs: dict

    onto_id_suffix: str
    onto_original_suffix: str

    load_raw: str
    mapped_features: str
    remove_gene_version: str

    ontology_constrained: List[str]

    obs_keys: List[str]
    var_keys: List[str]
    uns_keys: List[str]
    batch_keys: List[str]

    classmap_source_key: str
    classmap_target_key: str
    classmap_target_id_key: str

    invalid_metadata_identifier: Union[str, None]
    not_a_cell_celltype_identifier: Union[str, None]
    unknown_metadata_identifier: Union[str, None]

    @property
    def controlled_meta_keys(self):
        return [getattr(self, k) for k in self.obs_keys + self.uns_keys]

    @property
    def controlled_meta_fields(self):
        return [k for k in self.obs_keys + self.uns_keys]


class AdataIdsSfaira(AdataIds):
    """
    Class of constant field names of anndata.AnnData object entries, such as .uns keys and .obs columns in sfaira
    dataloader objects.
    """

    assay_differentiation: str
    assay_type_differentiation: str
    bio_sample: str
    cell_line: str

    def __init__(self):
        self.onto_id_suffix = "_ontology_term_id"
        self.onto_original_suffix = "_original"

        self.feature_kwargs = {
            "match_to_release": "104",
            "remove_gene_version": True,
            "subset_genes_to_type": None}

        self.annotated = "annotated"
        self.assay_sc = "assay_sc"
        self.assay_differentiation = "assay_differentiation"
        self.assay_type_differentiation = "assay_type_differentiation"
        self.author = "author"
        self.bio_sample = "bio_sample"
        self.cell_line = "cell_line"
        self.cell_type = "cell_type"
        self.default_embedding = "default_embedding"
        self.disease = "disease"
        self.doi_journal = "doi_journal"
        self.doi_preprint = "doi_preprint"
        self.dataset = "dataset"
        self.dataset_group = "dataset_group"
        self.download_url_data = "download_url_data"
        self.download_url_meta = "download_url_meta"
        self.feature_id = "ensembl"
        self.feature_index = self.feature_id
        self.feature_symbol = "gene_symbol"
        self.feature_biotype = "feature_biotype"
        self.id = "id"
        self.individual = "individual"
        self.ncells = "ncells"
        self.normalization = "normalization"
        self.organ = "organ"
        self.organism = "organism"
        self.primary_data = "primary_data"
        self.sample_source = "sample_source"
        self.tech_sample = "tech_sample"
        self.title = "title"
        self.year = "year"

        self.development_stage = "development_stage"
        self.ethnicity = "ethnicity"
        self.sex = "sex"
        self.state_exact = "state_exact"

        self.load_raw = "load_raw"
        self.mapped_features = "mapped_features"
        self.remove_gene_version = "remove_gene_version"

        self.classmap_source_key = "source"
        self.classmap_target_key = "target"
        self.classmap_target_id_key = "target_id"

        self.invalid_metadata_identifier = "na"
        self.not_a_cell_celltype_identifier = "NOT_A_CELL"
        self.unknown_metadata_identifier = "unknown"

        self.batch_keys = [self.bio_sample, self.individual, self.tech_sample]

        self.ontology_constrained = [
            "assay_sc",
            "cell_line",
            "cell_type",
            "development_stage",
            "disease",
            "ethnicity",
            "organ",
            "organism",
            "sex",
        ]
        self.obs_keys = [
            "assay_sc",
            "assay_differentiation",
            "assay_type_differentiation",
            "bio_sample",
            "cell_line",
            "cell_type",
            "development_stage",
            "disease",
            "ethnicity",
            "id",
            "individual",
            "organ",
            "organism",
            "sex",
            "state_exact",
            "sample_source",
            "tech_sample",
        ]
        self.var_keys = [
            "feature_id",
            "feature_symbol",
        ]
        self.uns_keys = [
            "annotated",
            "author",
            "default_embedding",
            "doi_journal",
            "doi_preprint",
            "download_url_data",
            "download_url_meta",
            "normalization",
            "primary_data",
            "title",
            "year",
            "load_raw",
            "mapped_features",
            "remove_gene_version",
        ] + [x for x in self.obs_keys if x not in self.batch_keys]


class AdataIdsCellxgene(AdataIds):
    """
    Class of constant field names of anndata.AnnData object entries", such as .uns keys and .obs columns in cellxgene
    objects.
    """
    accepted_file_names: List[str]

    def __init__(self):
        self.onto_id_suffix = "_ontology_term_id"
        self.onto_original_suffix = "_original"

        self.feature_kwargs = {
            "remove_gene_version": True,
            "subset_genes_to_type": None}

        self.assay_sc = "assay"
        self.author = "contributor"
        self.cell_type = "cell_type"
        self.default_embedding = "default_embedding"
        self.doi_journal = "publication_doi"
        self.doi_preprint = "preprint_doi"
        self.disease = "disease"
        self.feature_biotype = "feature_biotype"
        self.feature_id = "feature_id"
        self.feature_is_filtered = "feature_is_filtered"
        self.feature_reference = "feature_reference"
        self.feature_symbol = "feature_name"
        self.id = "id"
        self.ncells = "ncells"
        self.organ = "tissue"
        self.organism = "organism"
        self.primary_data = "is_primary_data"
        self.title = "title"
        self.year = "year"

        self.author = "contributors"
        self.development_stage = "development_stage"
        self.ethnicity = "ethnicity"
        self.sex = "sex"
        self.state_exact = "disease"
        self.tech_sample = "batch"

        # selected element entries used for parsing:
        self.author_names = "names"

        self.invalid_metadata_identifier = "na"
        self.not_a_cell_celltype_identifier = "CL:0000003"
        self.unknown_metadata_identifier = "unknown"

        self.batch_keys = []

        self.ontology_constrained = [
            "assay_sc",
            "cell_type",
            "development_stage",
            "disease",
            "ethnicity",
            "organ",
            "organism",
            "sex",
        ]

        self.obs_keys = [
            "assay_sc",
            "cell_type",
            "development_stage",
            "disease",
            "ethnicity",
            "organ",
            "organism",
            "primary_data",
            "sex",
            "tech_sample",
        ]
        self.var_keys = [
            "feature_id",
            "feature_symbol",
        ]
        self.uns_keys = [
            "doi_journal",
            "doi_preprint",
            "default_embedding",
            "title",
        ]
        # These attributes related to obs and uns keys above are also in the data set attributes that can be
        # inquired before download via the REST API:
        self.dataset_keys = [
            "assay_sc",
            "development_stage",
            "disease",
            "ethnicity",
            "organ",
            "organism",
            "sex",
        ]

    @property
    def feature_index(self):
        return self.feature_id


class AdataIdsCellxgene_v2_0_0(AdataIdsCellxgene):

    """
    https://github.com/chanzuckerberg/single-cell-curation/blob/main/schema/2.0.0/corpora_schema.md
    """

    def __init__(self):
        super(AdataIdsCellxgene_v2_0_0, self).__init__()
        self.feature_kwargs["match_to_release"] = "104"



import os
from typing import List, Union

from sfaira.data.store.stores.multi import StoresDao, StoresH5ad, \
    StoreMultipleFeatureSpaceBase


def load_store(cache_path: Union[str, os.PathLike, List[str], List[os.PathLike]], store_format: str = "dao",
               columns: Union[None, List[str]] = None) -> StoreMultipleFeatureSpaceBase:
    """
    Instantiates a distributed store class.

    Note that any store is instantiated as a DistributedStoreMultipleFeatureSpaceBase.
    This instances can be subsetted to the desired single feature space.

    :param cache_path: Store directory.
    :param store_format: Format of store {"h5ad", "dao"}.

        - "h5ad": Returns instance of DistributedStoreH5ad and keeps data in memory. See also "h5ad_backed".
        - "dao": Returns instance of DistributedStoreDoa (distributed access optimized).
        - "h5ad_backed": Returns instance of DistributedStoreH5ad and keeps data as backed (out of memory). See also
            "h5ad".
    :param columns: Which columns to read into the obs copy in the output, see pandas.read_parquet().
        Only relevant if store_format is "dao".
    :return: Instances of a distributed store class.
    """
    if store_format == "h5ad":
        return StoresH5ad(cache_path=cache_path, in_memory=True)
    elif store_format == "dao":
        return StoresDao(cache_path=cache_path, columns=columns)
    elif store_format == "h5ad_backed":
        return StoresH5ad(cache_path=cache_path, in_memory=False)
    else:
        raise ValueError(f"Did not recognize store_format {store_format}.")



import abc
import anndata
import numpy as np
import pandas as pd
import pickle
from typing import Union

from sfaira.consts import AdataIdsSfaira
from sfaira.data.store.stores.base import StoreBase
from sfaira.data import StoreSingleFeatureSpace, Universe
from sfaira.estimators.keras.base import EstimatorKeras, EstimatorKerasCelltype, EstimatorKerasEmbedding
from sfaira.ui import ModelZoo


class TrainModel:

    data: Union[anndata.AnnData, StoreSingleFeatureSpace]
    estimator: Union[EstimatorKeras, None]

    def __init__(
            self,
            data: Union[str, anndata.AnnData, Universe, StoreSingleFeatureSpace],
    ):
        # Check if handling backed anndata or base path to directory of raw files:
        if isinstance(data, str) and data.split(".")[-1] == "h5ad":
            self.data = anndata.read(data, backed='r')
            if len(self.data.obs.columns) == 0:
                fn_backed_obs = ".".join(data.split(".")[:-1]) + "_obs.csv"
                self.data.obs = pd.read_csv(fn_backed_obs)
        elif isinstance(data, anndata.AnnData):
            self.data = data
        elif isinstance(data, list) and isinstance(data[0], anndata.AnnData):
            self.data = data
        elif isinstance(data, Universe):
            self.data = data.adata
        elif isinstance(data, StoreBase):
            self.data = data
        else:
            raise ValueError(f"did not recognize data of type {type(data)}")
        self.zoo = ModelZoo()
        self._adata_ids = AdataIdsSfaira()

    def load_into_memory(self):
        """
        Loads backed objects from DistributedStoreBase into single adata object in memory in .data slot.
        :return:
        """
        if isinstance(self.data, StoreSingleFeatureSpace):
            adata = None
            for k, v in self.data.indices.items():
                x = self.data.adata_by_key[k][v, :].to_memory()
                x.obs["dataset_id"] = k
                if adata is None:
                    adata = x
                else:
                    adata = adata.concatenate(x)
            self.data = adata

    @property
    @abc.abstractmethod
    def topology_dict(self) -> dict:
        pass

    @abc.abstractmethod
    def init_estim(self):
        pass

    @abc.abstractmethod
    def save_eval(self, fn: str, **kwargs):
        pass

    @abc.abstractmethod
    def _save_specific(self, fn: str, **kwargs):
        pass

    def save(
            self,
            fn: str,
            model: bool = True,
            specific: bool = True
    ):
        """
        Save weights and summary statistics.
        """
        assert self.estimator is not None, "initialize estimator first"
        if model:
            self.estimator.model.training_model.save_weights(fn + "_weights.h5")
        self.save_eval(fn=fn)
        with open(fn + '_history.pickle', 'wb') as f:
            pickle.dump(obj=self.estimator.history, file=f)
        with open(fn + '_hyperparam.pickle', 'wb') as f:
            pickle.dump(obj=self.estimator.train_hyperparam, file=f)
        with open(fn + '_model_hyperparam.pickle', 'wb') as f:
            pickle.dump(obj=self.estimator.model.hyperparam, file=f)
        if specific:
            self._save_specific(fn=fn)

    def n_counts(self, idx):
        return np.asarray(self.estimator.data.checkout(idx=idx, return_dense=False).x.sum(axis=1)).flatten()


class TrainModelEmbedding(TrainModel):

    estimator: EstimatorKerasEmbedding

    def __init__(
            self,
            model_path: str,
            data: Union[str, anndata.AnnData, Universe, StoreSingleFeatureSpace],
    ):
        super(TrainModelEmbedding, self).__init__(data=data)
        self.estimator = None
        self.model_dir = model_path

    @property
    def topology_dict(self) -> dict:
        topology_dict = self.zoo.topology_container.topology
        return topology_dict

    def init_estim(
            self,
            override_hyperpar: Union[dict, None] = None
    ):
        assert self.zoo.model_id is not None, "choose model in zoo first"
        self.estimator = EstimatorKerasEmbedding(
            data=self.data,
            model_dir=self.model_dir,
            model_id=self.zoo.model_id,
            model_topology=self.zoo.topology_container,
            cache_path=self.model_dir,
        )
        self.estimator.init_model(override_hyperpar=override_hyperpar)
        print(f"TRAINER: initialised model with {self.estimator.topology_container.n_var} features.")

    def save_eval(self, fn: str, **kwargs):
        evaluation_train = self.estimator.evaluate_any(idx=self.estimator.idx_train)
        evaluation_val = self.estimator.evaluate_any(idx=self.estimator.idx_eval)
        evaluation_test = self.estimator.evaluate_any(idx=self.estimator.idx_test)
        evaluation_all = self.estimator.evaluate_any(idx=None)
        evaluation = {
            'test': evaluation_test,
            'val': evaluation_val,
            'train': evaluation_train,
            'all': evaluation_all
        }
        with open(fn + '_evaluation.pickle', 'wb') as f:
            pickle.dump(obj=evaluation, file=f)

    def _save_specific(self, fn: str, **kwargs):
        """
        Save embedding prediction:

        :param fn:
        :return:
        """
        embedding = self.estimator.predict_embedding()
        df_summary = self.estimator.obs_test
        df_summary = df_summary[[k for k in df_summary.columns if k in self._adata_ids.obs_keys]]
        df_summary["ncounts"] = self.n_counts(idx=self.estimator.idx_test)
        np.save(file=fn + "_embedding", arr=embedding)
        df_summary.to_csv(fn + "_covar.csv")
        with open(fn + "_topology.pickle", "wb") as f:
            pickle.dump(obj=self.topology_dict, file=f)


class TrainModelCelltype(TrainModel):

    estimator: EstimatorKerasCelltype

    def __init__(
            self,
            model_path: str,
            data: Union[str, anndata.AnnData, Universe, StoreSingleFeatureSpace],
            fn_target_universe: str,
    ):
        super(TrainModelCelltype, self).__init__(data=data)
        self.estimator = None
        self.model_dir = model_path
        self.fn_target_universe = fn_target_universe

    @property
    def topology_dict(self) -> dict:
        topology_dict = self.zoo.topology_container.topology
        # Load target universe leaves into topology dict:
        topology_dict["output"]["targets"] = self.estimator.celltype_universe.onto_cl.leaves
        return topology_dict

    def init_estim(
            self,
            override_hyperpar: Union[dict, None] = None
    ):
        assert self.zoo.model_id is not None, "choose model in zoo first"
        self.estimator = EstimatorKerasCelltype(
            data=self.data,
            model_dir=self.model_dir,
            model_id=self.zoo.model_id,
            model_topology=self.zoo.topology_container,
            cache_path=self.model_dir,
        )
        self.estimator.celltype_universe.load_target_universe(self.fn_target_universe)
        self.estimator.init_model(override_hyperpar=override_hyperpar)
        print(f"TRAINER: initialised model with {self.estimator.topology_container.n_var} features and "
              f"{self.estimator.ntypes} labels: \n{self.estimator.ontology_names}.")

    def save_eval(self, fn: str, eval_weighted: bool = False, **kwargs):
        evaluation = {
            'train': self.estimator.evaluate_any(idx=self.estimator.idx_train, weighted=False),
            'val': self.estimator.evaluate_any(idx=self.estimator.idx_eval, weighted=False),
            'test': self.estimator.evaluate_any(idx=self.estimator.idx_test, weighted=False),
            'all': self.estimator.evaluate_any(idx=None, weighted=False)
        }
        with open(fn + '_evaluation.pickle', 'wb') as f:
            pickle.dump(obj=evaluation, file=f)
        if eval_weighted:
            evaluation_weighted = {
                'train': self.estimator.evaluate_any(idx=self.estimator.idx_train, weighted=True),
                'val': self.estimator.evaluate_any(idx=self.estimator.idx_eval, weighted=True),
                'test': self.estimator.evaluate_any(idx=self.estimator.idx_test, weighted=True),
                'all': self.estimator.evaluate_any(idx=None, weighted=True)
            }
            with open(fn + '_evaluation_weighted.pickle', 'wb') as f:
                pickle.dump(obj=evaluation_weighted, file=f)

    def _save_specific(self, fn: str, **kwargs):
        """
        Save true and predicted labels on test set:

        :param fn:
        :return:
        """
        ytrue = self.estimator.ytrue()
        yhat = self.estimator.predict()
        df_summary = self.estimator.obs_test
        df_summary = df_summary[[k for k in df_summary.columns if k in self._adata_ids.obs_keys]]
        df_summary["ncounts"] = self.n_counts(idx=self.estimator.idx_test)
        np.save(file=fn + "_ytrue", arr=ytrue)
        np.save(file=fn + "_yhat", arr=yhat)
        df_summary.to_csv(fn + "_covar.csv")
        with open(fn + '_ontology_names.pickle', 'wb') as f:
            pickle.dump(obj=self.estimator.ontology_names, file=f)
        with open(fn + '_ontology_ids.pickle', 'wb') as f:
            pickle.dump(obj=self.estimator.ontology_ids, file=f)
        with open(fn + "_topology.pickle", "wb") as f:
            pickle.dump(obj=self.topology_dict, file=f)

        obs = self.estimator.data.checkout(idx=None, obs_keys=["cell_type"]).obs
        cell_counts = obs['cell_type'].value_counts().to_dict()
        with open(fn + '_celltypes_valuecounts_wholedata.pickle', 'wb') as f:
            pickle.dump(obj=[cell_counts], file=f)



import tarfile

import anndata
import os

import pandas as pd
import scipy.sparse
import h5py


def load(data_dir, sample_fn, **kwargs):
    fn = os.path.join(data_dir, 'GSE122960_RAW.tar')
    with tarfile.open(fn) as tar:
        f = h5py.File(tar.extractfile(f'{sample_fn}_filtered_gene_bc_matrices_h5.h5'), 'r')['GRCh38']
        x = scipy.sparse.csc_matrix((f['data'], f['indices'], f['indptr']), shape=f['shape']).T
        var = pd.DataFrame({'feature_id': f['genes'], 'feature_symbol': f['gene_names']})
        obs = pd.DataFrame({'barcode': f['barcodes']})
        adata = anndata.AnnData(X=x, obs=obs, var=var)

    return adata



# -*- coding: utf-8 -*-
"""A Data and Model Zoo for Single-Cell Genomics."""

from ._settings import settings
import sfaira.consts
import sfaira.data
import sfaira.models
import sfaira.train
import sfaira.ui
import sfaira.versions
from ._version import get_versions

__version__ = get_versions()['version']
del get_versions
__maintainer__ = ', '.join([
    "Leander Dony",
    "David S. Fischer"
])
__author__ = ', '.join([
    "Leander Dony",
    "David S. Fischer",
    "Lukas Heumos"
])
__email__ = ', '.join([
    "leander.dony@helmholtz-munich.de",
    "david.fischer@helmholtz-munich.de",
    "lukas.heumos@helmholtz-munich.de"
])



import anndata
import os

from sfaira.data import DatasetBase

SAMPLE_FNS = [
    "tabula-muris-senis-droplet-processed-official-annotations-Fat.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-BAT.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-GAT.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-MAT.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-SCAT.h5ad",
    "tabula-muris-senis-droplet-processed-official-annotations-Bladder.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Bladder.h5ad",
    "tabula-muris-senis-droplet-processed-official-annotations-Marrow.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Marrow.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Brain_Non-Myeloid.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Brain_Myeloid.h5ad",
    "tabula-muris-senis-droplet-processed-official-annotations-Large_Intestine.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Large_Intestine.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Diaphragm.h5ad",
    "tabula-muris-senis-droplet-processed-official-annotations-Heart_and_Aorta.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Heart.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Aorta.h5ad",
    "tabula-muris-senis-droplet-processed-official-annotations-Kidney.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Kidney.h5ad",
    "tabula-muris-senis-droplet-processed-official-annotations-Liver.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Liver.h5ad",
    "tabula-muris-senis-droplet-processed-official-annotations-Lung.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Lung.h5ad",
    "tabula-muris-senis-droplet-processed-official-annotations-Mammary_Gland.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Mammary_Gland.h5ad",
    "tabula-muris-senis-droplet-processed-official-annotations-Limb_Muscle.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Limb_Muscle.h5ad",
    "tabula-muris-senis-droplet-processed-official-annotations-Pancreas.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Pancreas.h5ad",
    "tabula-muris-senis-droplet-processed-official-annotations-Skin.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Skin.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Spleen.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Spleen.h5ad",
    "tabula-muris-senis-droplet-processed-official-annotations-Thymus.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Thymus.h5ad",
    "tabula-muris-senis-droplet-processed-official-annotations-Tongue.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Tongue.h5ad",
    "tabula-muris-senis-droplet-processed-official-annotations-Trachea.h5ad",
    "tabula-muris-senis-facs-processed-official-annotations-Trachea.h5ad",
]


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        organ = "-".join(self.sample_fn.split("-")[7:]).split(".")[0].lower()
        organ = "adipose tissue" if organ in ["fat", "bat", "gat", "mat", "scat"] else \
            "aorta" if organ in ["aorta"] else \
            "urinary bladder" if organ in ["bladder"] else \
            "bone marrow" if organ in ["marrow"] else \
            "brain" if organ in ["brain_non-myeloid", "brain_myeloid"] else \
            "colon" if organ in ["large_intestine"] else \
            "diaphragm" if organ in ["diaphragm"] else \
            "heart" if organ in ["heart_and_aorta", "heart"] else \
            "kidney" if organ in ["kidney"] else \
            "liver" if organ in ["liver"] else \
            "lung" if organ in ["lung"] else \
            "mammary gland" if organ in ["mammary_gland"] else \
            "muscle organ" if organ in ["limb_muscle"] else \
            "pancreas" if organ in ["pancreas"] else \
            "skin of body" if organ in ["skin"] else \
            "spleen" if organ in ["spleen"] else \
            "thymus" if organ in ["thymus"] else \
            "tongue" if organ in ["tongue"] else \
            "trachea" if organ in ["trachea"] else organ
        # ToDo: heart_and_aorta could be a distinct UBERON term, e.g. cardiovascular system?

        self.download_url_data = f"https://czb-tabula-muris-senis.s3-us-west-2.amazonaws.com/Data-objects/" \
                                 f"{self.sample_fn}"
        self.download_url_meta = None

        self.cell_type_obs_key = "cell_ontology_class"
        self.development_stage_obs_key = "development_stage"
        self.sex_obs_key = "sex"
        # ToDo: further anatomical information for subtissue in "subtissue"?

        self.author = "Pisco"
        self.disease = "healthy"
        self.doi_journal = "10.1038/s41586-020-2496-1"
        self.doi_preprint = "10.1101/661728"
        self.layer_processed = "X"
        self.organism = "Mus musculus"
        self.organ = organ
        self.primary_data = True
        self.assay_sc = "10x 3' v2" if self.sample_fn.split("-")[3] == "droplet" else "Smart-seq2"
        self.year = 2019
        self.sample_source = "primary_tissue"

        self.feature_id_var_key = None
        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"

        self.set_dataset_id(idx=1)


def load(data_dir, sample_fn, **kwargs):
    dev_stage_dict = {
        "18m": "18 month-old stage",
        "1m": "4 weeks",
        "21m": "20 month-old stage and over",
        "24m": "20 month-old stage and over",
        "30m": "20 month-old stage and over",
        "3m": "3 month-old stage",
    }
    fn = os.path.join(data_dir, sample_fn)
    adata = anndata.read_h5ad(fn)
    adata.X = adata.raw.X
    adata.var = adata.raw.var
    del adata.raw
    adata.obsm = {}
    adata.varm = {}
    adata.uns = {}
    adata.obs['development_stage'] = [dev_stage_dict[i] for i in adata.obs['age']]

    return adata



import os
import pandas
import shutil
import zipfile


def load(data_dir, **kwargs):
    age_dict = {
        17: "17th week post-fertilization homosapiens stage",
        18: "18th week post-fertilization homosapiens stage",
    }
    ct_dict = {
        "End": "Endothelial",
        "ExDp1": "Excitatory deep layer 1",
        "ExDp2": "Excitatory deep layer 2",
        "ExM": "Maturing excitatory",
        "ExM-U": "Maturing excitatory upper enriched",
        "ExN": "Migrating excitatory",
        "IP": "IP",
        "InCGE": "Interneuron CGE",
        "InMGE": "Interneuron MGE",
        "Mic": "Microglia",
        "OPC": "OPC",
        "Per": "Pericyte",
        "PgG2M": "Cycling Progenitors (G2/M phase)",
        "PgS": "Cycling Progenitors (S phase)",
        "oRG": "Outer Radial Glia",
        "vRG": "Ventricular Radial Glia",
    }

    import anndata2ri
    from rpy2.robjects import r
    fn = os.path.join(data_dir, "sc_dev_cortex_geschwind.zip")
    fn_tmp = os.path.join(os.path.expanduser("~"), "sfaira_tmp")
    if not os.path.exists(fn_tmp):
        os.makedirs(fn_tmp)
    with zipfile.ZipFile(fn, 'r') as zip_ref:
        zip_ref.extractall(fn_tmp)
    anndata2ri.activate()  # TODO: remove global activation of anndata2ri and use localconverter once it's fixed
    adata = r(
        f"library(Seurat)\n"
        f"load('{os.path.join(fn_tmp, 'raw_counts_mat.rdata')}')\n"
        f"new_obj = CreateSeuratObject(raw_counts_mat)\n"
        f"as.SingleCellExperiment(new_obj)\n"
    )
    obs = pandas.read_csv(os.path.join(fn_tmp, "cell_metadata.csv"), index_col=0)
    adata = adata[obs.index.tolist()].copy()
    adata.obs = obs
    shutil.rmtree(fn_tmp)
    adata.obs['devstage'] = [age_dict[i] for i in adata.obs['Gestation_week']]
    adata.obs['celltype'] = [ct_dict[i] for i in adata.obs['Cluster']]
    return adata



import pydoc
import os
from typing import List
from warnings import warn
from sfaira.data import DatasetBase, DatasetGroup, DatasetSuperGroup, DatasetGroupDirectoryOriented

from sfaira.unit_tests.directories import DIR_DATA_LOADERS_CACHE


class DatasetSuperGroupMock(DatasetSuperGroup):
    """
    This is a DatasetSuperGroup which wraps the mock data loaders in the same directory.

    This class is designed to facilitate testing of code that requires data loaders without requiring raw data
    downloads as all mock data loaders operate on data that is simulated in the `load()` functions.
    A cache directory is established under ../cache.

    This class is a reduced and merged version of the sfaira loader super group class and the sfaira loader adapated
    DatasetGroupDirectoryOriented.
    """

    dataset_groups: List[DatasetGroupDirectoryOriented]

    def __init__(self):
        # Directory choice hyper-paramters:
        dir_prefix = "d"
        dir_exclude = []
        # Collect all data loaders from files in directory:
        dataset_groups = []
        cwd = os.path.dirname(__file__)
        for d in os.listdir(cwd):
            if os.path.isdir(os.path.join(cwd, d)):  # Iterate over mock studies (directories).
                if d[:len(dir_prefix)] == dir_prefix and d not in dir_exclude:  # Narrow down to data set directories
                    path_base = f"sfaira.unit_tests.data_for_tests.loaders.loaders.{d}"
                    path_dsg = pydoc.locate(f"{path_base}.FILE_PATH")
                    path_module = os.path.join(cwd, d)
                    for f in os.listdir(os.path.join(cwd, d)):  # Iterate over loaders in mock study (file).
                        datasets = []
                        if f.split(".")[-1] == "py" and f not in ["__init__.py"]:
                            file_module = ".".join(f.split(".")[:-1])
                            if path_dsg is not None:
                                load_func = pydoc.locate(f"{path_base}.{file_module}.load")
                                fn_yaml = os.path.join(path_module, file_module + ".yaml")
                                x = DatasetBase(
                                    data_path=DIR_DATA_LOADERS_CACHE,
                                    meta_path=DIR_DATA_LOADERS_CACHE,
                                    cache_path=DIR_DATA_LOADERS_CACHE,
                                    load_func=load_func,
                                    dict_load_func_annotation=None,
                                    sample_fn=None,
                                    sample_fns=None,
                                    yaml_path=fn_yaml,
                                )
                                x.read_ontology_class_maps(fns=[os.path.join(path_module, file_module + ".tsv")])
                                datasets.append(x)
                            else:
                                warn(f"DatasetGroupDirectoryOriented was None for {f}")
                        dataset_groups.append(DatasetGroup(datasets=dict([(x.id, x) for x in datasets]),
                                                           collection_id=d))
        super().__init__(dataset_groups=dataset_groups)



import os
import scipy.sparse
import anndata as ad
import numpy as np
import tables

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = \
            "http://cf.10xgenomics.com/samples/cell-exp/3.0.0/pbmc_10k_v3/pbmc_10k_v3_filtered_feature_bc_matrix.h5"
        self.download_url_meta = None

        self.assay_sc = "10x 3' v3"
        self.author = "10x Genomics"
        self.disease = "healthy"
        self.doi_journal = "no_doi_10x_genomics"
        self.layer_counts = "X"
        self.organ = "blood"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.year = 2019

        self.feature_id_var_key = "gene_ids"
        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = os.path.join(data_dir, "pbmc_10k_v3_filtered_feature_bc_matrix.h5")
    with tables.open_file(str(fn), 'r') as f:
        dsets = {}
        for node in f.walk_nodes('/matrix', 'Array'):
            dsets[node.name] = node.read()

    M, N = dsets['shape']
    data = dsets['data']
    if dsets['data'].dtype == np.dtype('int32'):
        data = dsets['data'].view('float32')
        data[:] = dsets['data']
    matrix = scipy.sparse.csr_matrix(
        (data, dsets['indices'], dsets['indptr']),
        shape=(N, M),
    )
    adata = ad.AnnData(
        matrix,
        dict(obs_names=dsets['barcodes'].astype(str)),
        dict(
            var_names=dsets['name'].astype(str),
            gene_ids=dsets['id'].astype(str),
            feature_types=dsets['feature_type'].astype(str),
            genome=dsets['genome'].astype(str),
        ),
    )

    return adata



import anndata
import os
import scipy.sparse


def load(data_dir, sample_fn, **kwargs):
    fn = os.path.join(data_dir, sample_fn)
    adata = anndata.read(fn)
    adata.X = scipy.sparse.csr_matrix(adata.X)
    adata.obs["age"] = [f"{x}-year-old human stage" for x in adata.obs["age"].values]

    return adata



import anndata
import gzip
import os
import pandas as pd
import scipy.io
import tarfile


def load(data_dir, sample_fn, **kwargs):
    fn = os.path.join(data_dir, "GSE164690_RAW.tar")
    with tarfile.open(fn) as tar:
        with gzip.open(tar.extractfile(f"{sample_fn}_matrix.mtx.gz"), "rb") as mm:
            x = scipy.io.mmread(mm).T.tocsr()
        obs = pd.read_csv(tar.extractfile(f"{sample_fn}_barcodes.tsv.gz"), compression="gzip", header=None,
                          sep="\t", index_col=0)
        obs.index.name = None
        var = pd.read_csv(tar.extractfile(f"{sample_fn}_features.tsv.gz"), compression="gzip", header=None, sep="\t")
        var.columns = ["ensembl", "symbol", "feature_class"]
        var.index = var["ensembl"].values
        adata = anndata.AnnData(X=x, obs=obs, var=var)
    # Annotate organism in .var
    adata.var["organism"] = ["Homo sapiens" if x.startswith("ENSG") else "Human papillomavirus"
                             for x in adata.var["ensembl"].values]
    # TODO: allow HPV back in once multi-organism is supported.
    adata = adata[:, adata.var["organism"].values == "Homo sapiens"].copy()
    adata.obs["patient"] = sample_fn.split("_")[1]
    return adata



import anndata
import os
import numpy as np

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    """
    TODO transform field development to controlled field age
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = [
            "https://cellgeni.cog.sanger.ac.uk/BenKidney_v2.1/Mature_Full_v2.1.h5ad",
            "https://cellgeni.cog.sanger.ac.uk/BenKidney_v2.1/Fetal_full.h5ad"
        ]
        self.download_url_meta = None

        self.assay_sc = "10x 3' v2"
        self.author = "Stewart"
        self.disease = "healthy"
        self.doi_journal = "10.1126/science.aat5031"
        self.layer_processed = "X"
        self.organ = "kidney"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.state_exact = "healthy"
        self.year = 2019

        self.feature_id_var_key = "ID"
        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"
        self.cell_type_obs_key = "celltype"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = [
        os.path.join(data_dir, "Mature_Full_v2.1.h5ad"),
        os.path.join(data_dir, "Fetal_full.h5ad")
    ]
    adult = anndata.read(fn[0])
    fetal = anndata.read(fn[1])
    # TODO this is is not a controlled field
    adult.obs["development"] = "adult"
    fetal.obs["development"] = "fetal"
    adata = adult.concatenate(fetal)
    adata.X = np.expm1(adata.X)

    return adata



import numpy as np
try:
    import tensorflow as tf
except ImportError:
    tf = None


class LossLoglikelihoodNb(tf.keras.losses.Loss):

    def __init__(self, average=True, *args, **kwargs):
        super(LossLoglikelihoodNb, self).__init__(*args, **kwargs, name="ll_nb")
        self.average = average

    def call(
            self,
            y_true,
            y_pred
    ):
        """Implements the negative log likelihood loss as VAE reconstruction loss"""
        x = y_true
        loc, scale = tf.split(y_pred, num_or_size_splits=2, axis=1)

        eta_loc = tf.math.log(loc)
        eta_scale = tf.math.log(scale)

        log_r_plus_mu = tf.math.log(scale + loc)

        ll = tf.math.lgamma(scale + x)
        ll = ll - tf.math.lgamma(x + tf.ones_like(x))
        ll = ll - tf.math.lgamma(scale)
        ll = ll + tf.multiply(x, eta_loc - log_r_plus_mu) + tf.multiply(scale, eta_scale - log_r_plus_mu)

        ll = tf.clip_by_value(ll, -300, 300, "log_probs")
        neg_ll = -ll
        if self.average:
            neg_ll = tf.reduce_mean(neg_ll)
        else:
            # sum over features, average over batch
            neg_ll = tf.reduce_mean(tf.reduce_sum(neg_ll, axis=1), axis=0)
        return neg_ll


class LossLoglikelihoodGaussian(tf.keras.losses.Loss):

    def __init__(self, average=True, *args, **kwargs):
        super(LossLoglikelihoodGaussian, self).__init__(*args, **kwargs, name="ll_norm")
        self.average = average

    def call(
            self,
            y_true,
            y_pred
    ):
        """Implements the gaussian log likelihood loss as VAE reconstruction loss"""
        loc, scale = tf.split(y_pred, num_or_size_splits=2, axis=1)

        ll = -tf.math.log(scale * tf.math.sqrt(2. * np.pi)) - 0.5 * tf.math.square((y_true - loc) / scale)
        ll = tf.clip_by_value(ll, -300, 300, "log_probs")
        neg_ll = -ll
        if self.average:
            neg_ll = tf.reduce_mean(neg_ll)
        else:
            # sum over features, average over batch
            neg_ll = tf.reduce_mean(tf.reduce_sum(neg_ll, axis=1), axis=0)
        return neg_ll


class LossCrossentropyAgg(tf.keras.losses.Loss):

    def __init__(self, *args, **kwargs):
        super(LossCrossentropyAgg, self).__init__(*args, **kwargs, name="cce_agg")

    def call(
            self,
            y_true,
            y_pred
    ):
        """ Modified crossentropy that aggregates allowed output classes into single class. """
        y_pred = tf.clip_by_value(y_pred, 1e-10, 1., "y_pred")
        ll_cce_agg = -tf.math.log(tf.reduce_mean(y_true * y_pred, axis=1, keepdims=False))
        return ll_cce_agg


class KLLoss(tf.keras.losses.Loss):

    def __init__(self):
        super(KLLoss, self).__init__()
        self.beta = tf.Variable(1.0, dtype=tf.float32, trainable=False)

    def call(
            self,
            y_true,
            y_pred
    ):
        expected_logqz_x, expected_logpz = tf.split(y_pred, num_or_size_splits=2, axis=1)

        kl_loss = tf.reduce_mean(expected_logqz_x - expected_logpz, axis=0)
        return self.beta * kl_loss



import sys
import logging
import os
import pydoc
import rich
from rich.panel import Panel
from rich.progress import Progress, BarColumn
from typing import Dict, List
import yaml

from sfaira.commands.utils import get_pydoc
from sfaira.consts.utils import clean_doi

log = logging.getLogger(__name__)


class DataloaderValidator:
    fns_loaders: List[str]
    paths_yamls: List[str]
    yaml_dicts: Dict[str, dict]
    passed: List[str]
    warned: List[str]
    failed: List[str]

    def __init__(self, path_loader, doi, schema):
        self.doi = doi
        self._clean_doi = clean_doi(doi)

        loader_filenames = [str(x) for x in os.listdir(os.path.join(path_loader, self._clean_doi))
                            if str(x).endswith(".py") and str(x) != "__init__.py"]
        yaml_filenames = [x for x in os.listdir(os.path.join(path_loader, self._clean_doi))
                          if str(x).endswith(".yaml")]
        self.path_loader = path_loader
        self.fns_loaders = loader_filenames
        self.paths_yamls = [os.path.join(path_loader, self._clean_doi, x) for x in yaml_filenames]
        self.yaml_dicts = {}
        self.passed = []
        self.warned = []
        self.failed = []
        self.loader_validation_functions: list = [
            '_validate_namespace',
        ]
        self.yaml_validation_functions: list = [
            '_validate_yaml_sections',
            '_validate_required_attributes',
        ]
        self.schema = schema

    def validate(self) -> None:
        """
        Statically verifies data loader files against a predefined set of rules.
        """
        progress = Progress("[bold green]{task.description}", BarColumn(bar_width=None),
                            "[bold yellow]{task.completed} of {task.total}[reset] [bold green]{task.fields[func_name]}")
        self._validate_loader(progress=progress)
        self._validate_yaml(progress=progress)
        self._print_results()

    def _validate_loader(self, progress) -> None:
        """
        Statically verifies a loader (.py) file against a predefined set of rules.
        Every rule is a function defined in this class, which must be part of this class' linting_functions.
        """
        with progress:
            lint_progress = progress.add_task("Running loader lint checks",
                                              total=len(self.yaml_validation_functions),
                                              func_name=self.yaml_validation_functions)
            for fun_name in self.loader_validation_functions:
                progress.update(lint_progress, advance=1, func_name=fun_name)
                getattr(self, fun_name)()

    def _validate_namespace(self):
        """
        Verifies that namespace of .py files contain all necessary elements.

            - .py files contain load() function.
        """
        passed_required_elements = True

        elements = ['load', ]

        for fn in self.fns_loaders:
            file_module = ".".join(fn.split(".")[:-1])
            _, pydoc_handle = get_pydoc(path_loader=self.path_loader, doi_sfaira_repr=self._clean_doi)
            pydoc_handle = pydoc_handle + "." + file_module
            for x in elements:
                query_element = pydoc.locate(pydoc_handle + "." + x)
                if query_element is None:
                    passed_required_elements = False
                    self.failed.append(f'Missing element in namespace of {fn}.py file: {x}')

        if passed_required_elements:
            self.passed.append('Passed required data loader load() namespace checks.')

    def _validate_yaml(self, progress) -> None:
        """
        Statically verifies the yaml data loader files against a predefined set of rules.
        Every rule is a function defined in this class, which must be part of this class' linting_functions.
        """
        for x in self.paths_yamls:
            with open(x) as yaml_file:
                self.yaml_dicts[x] = yaml.load(yaml_file, Loader=yaml.FullLoader)

        with progress:
            lint_progress = progress.add_task("Running yaml lint checks",
                                              total=len(self.yaml_validation_functions),
                                              func_name=self.yaml_validation_functions)
            for fun_name in self.yaml_validation_functions:
                progress.update(lint_progress, advance=1, func_name=fun_name)
                getattr(self, fun_name)()

    def _validate_yaml_sections(self):
        """
        Verifies that .yaml files contain correct sections and keys within each section.

        This can prevent copy or delete mistakes, for example.
        """
        passed_required_sections = True

        # Note: Each key in 'dataset_or_observation_wise' appears as written below and as f"{x}_obs_key" in the .yaml.
        # Each key is accordingly listed twice and the list of lists is flattened in this dictionary:
        section_keys = {
            'dataset_structure': [
                'dataset_index',
                'sample_fns'],
            'dataset_wise': [
                'author',
                'default_embedding',
                'doi_journal',
                'doi_preprint',
                'download_url_data',
                'download_url_meta',
                'primary_data',
                'year'],
            'layers': [
                'layer_counts',
                'layer_processed',
                'layer_spliced_counts',
                'layer_spliced_processed',
                'layer_unspliced_counts',
                'layer_unspliced_processed',
                'layer_velocity'],
            'dataset_or_feature_wise': [
                'feature_reference',
                'feature_reference_var_key',
                'feature_type',
                'feature_type_var_key'],
            'dataset_or_observation_wise': [z for y in [[x, x + "_obs_key"] for x in [
                'assay_sc',
                'assay_differentiation',
                'assay_type_differentiation',
                'bio_sample',
                'cell_line',
                'cell_type',
                'development_stage',
                'disease',
                'ethnicity',
                'gm',
                'individual',
                'organ',
                'organism',
                'sample_source',
                'sex',
                'source_doi',
                'state_exact',
                'tech_sample',
                'treatment']] for z in y],
            'feature_wise': [
                'feature_id_var_key',
                'feature_symbol_var_key'],
            'observation_wise': [
                'spatial_x_coord_obs_key',
                'spatial_y_coord_obs_key',
                'spatial_z_coord_obs_key',
                'vdj_vj_1_obs_key_prefix',
                'vdj_vj_2_obs_key_prefix',
                'vdj_vdj_1_obs_key_prefix',
                'vdj_vdj_2_obs_key_prefix',
                'vdj_c_call_obs_key_suffix',
                'vdj_consensus_count_obs_key_suffix',
                'vdj_d_call_obs_key_suffix',
                'vdj_duplicate_count_obs_key_suffix',
                'vdj_j_call_obs_key_suffix',
                'vdj_junction_obs_key_suffix',
                'vdj_junction_aa_obs_key_suffix',
                'vdj_locus_obs_key_suffix',
                'vdj_productive_obs_key_suffix',
                'vdj_v_call_obs_key_suffix'],
            'meta': [
                'version'],
        }

        for fn, content in self.yaml_dicts.items():
            for k, v in section_keys.items():
                if k not in content.keys():
                    passed_required_sections = False
                    self.failed.append(f'Missing section {k} in file {fn}')
                else:
                    for kk in v:
                        if kk not in content[k].keys():
                            passed_required_sections = False
                            self.failed.append(f'Missing key {kk} in section {k} in file {fn}')

        if passed_required_sections:
            self.passed.append('Passed required data loader section checks.')

    def _validate_required_attributes(self):
        """
        Verifies that all required attributes for every data loader are present.
        """
        passed_required_attributes = True

        attributes = {
            'dataset_structure': [
                'dataset_index'],
            'dataset_wise': [
                'author',
                ['doi_journal', 'doi_preprint'],
                'download_url_data',
                'primary_data',
                'year'],
            'layers': [
                ['layer_counts', 'layer_processed']],
            'dataset_or_feature_wise': [
                ['feature_type', 'feature_type_var_key']],
            'feature_wise': [
                ['feature_id_var_key', 'feature_symbol_var_key']],
            'meta': [
                'version'],
        }
        if self.schema == "sfaira":
            attributes['dataset_or_observation_wise'] = [[x, x + "_obs_key"] for x in [
                'assay_sc',
                'organism']]
        elif self.schema == "cellxgene":
            attributes['dataset_wise'].append("default_embedding")
            attributes['dataset_or_observation_wise'] = [[x, x + "_obs_key"] for x in [
                'assay_sc',
                'cell_type',
                'development_stage',
                'disease',
                'ethnicity',
                'organ',
                'organism',
                'sex']]
        else:
            print(f"[bold red]Did not recognize schema {self.schema} in validate.")
            sys.exit()

        for fn, yaml_dict in self.yaml_dicts.items():
            for section, attrs in attributes.items():
                for attr in attrs:
                    detected = False
                    # Lists of attributes are handled in the following way:
                    # One of the two keys has to have a value.
                    if isinstance(attr, list):
                        for sub_attr in attr:
                            val = yaml_dict[section][sub_attr]
                            if val is not None:
                                detected = True
                        if not detected:
                            passed_required_attributes = False
                            self.failed.append('Missing any of the following keys to describe the required meta ' +
                                               f'data {section}:{attr} in file {fn}')
                    else:
                        val = yaml_dict[section][attr]
                        if val is not None:
                            detected = True
                        if not detected:
                            passed_required_attributes = False
                            self.failed.append(f'Missing required meta data {section}:{attr} in file {fn}')

        if passed_required_attributes:
            self.passed.append('Passed required meta data checks.')

    def _print_results(self):
        console = rich.console.Console()
        console.print()
        console.rule("[bold green] LINT RESULTS")
        console.print()
        console.print(
            f'     [bold green][[\u2714]] {len(self.passed):>4} tests passed\n'
            f'     [bold yellow][[!]] {len(self.warned):>4} tests had warnings\n'
            f'     [bold red][[\u2717]] {len(self.failed):>4} tests failed',
            overflow="ellipsis",
            highlight=False,
        )

        def format_result(linting_results: dict, color):
            results = []
            for result in linting_results:
                results.append(f'[bold {color}]Result: {result}')
            return "\n".join(results)

        if len(self.passed) > 0:
            console.print()
            console.rule("[bold green][[\u2714]] Tests Passed", style='green')
            console.print(Panel(format_result(self.passed, 'green'), style='green'), no_wrap=False, overflow='ellipsis')
        if len(self.warned) > 0:
            console.print()
            console.rule("[bold yellow][[!]] Test Warnings", style='yellow')
            console.print(Panel(format_result(self.warned, 'yellow'), style="yellow"), no_wrap=False,
                          overflow='ellipsis')
        if len(self.failed) > 0:
            console.print()
            console.rule("[bold red][[\u2717]] Test Failures", style='red')
            console.print(Panel(format_result(self.failed, 'red'), style='red'), no_wrap=False, overflow='ellipsis')



import numpy as np
import pandas as pd
from typing import Dict, List, Tuple, Union

from sfaira.versions.metadata import OntologyCl, OntologyList, OntologyUberon

TARGET_UNIVERSE_KEY_NAME = "name"
TARGET_UNIVERSE_KEY_ID = "id"


class CelltypeUniverse:
    """
    Cell type universe (list) and ontology (hierarchy) container class.


    Basic checks on the organ specific instance are performed in the constructor.
    """
    onto_cl: OntologyCl
    onto_uberon: OntologyUberon
    _target_universe: Union[List[str], None]

    def __init__(self, cl: Union[OntologyCl, OntologyList], uberon: OntologyUberon, **kwargs):
        self.onto_cl = cl
        self.onto_uberon = uberon
        self._target_universe = None

    def __validate_target_universe_table(self, tab: pd.DataFrame):
        assert len(tab.columns) == 2
        assert tab.columns[0] == "name" and tab.columns[1] == "id"

    def load_target_universe(self, fn):
        """

        :param fn: .csv file containing target universe.
        :return:
        """
        tab = pd.read_csv(fn, sep="\t", index_col=None)
        self.__validate_target_universe_table(tab=tab)
        self.onto_cl.leaves = tab["name"].values

    def write_target_universe(
            self,
            fn,
            x: List[str],
    ):
        """

        :param fn: .csv file containing target universe.
        :param x: Nodes that make up target universe.
        :return:
        """
        tab = pd.DataFrame({
            TARGET_UNIVERSE_KEY_NAME: self.onto_cl.convert_to_name(x),
            TARGET_UNIVERSE_KEY_ID: self.onto_cl.convert_to_id(x),
        })
        self.__validate_target_universe_table(tab=tab)
        tab.to_csv(path_or_buf=fn, sep="\t", index=False)



try:
    import tensorflow as tf
except ImportError:
    tf = None


class NegBinOutput(tf.keras.layers.Layer):
    """Negative binomial output layer"""

    def __init__(
            self,
            original_dim=None,
            name='neg_bin_output',
            **kwargs
    ):

        super().__init__(name=name, **kwargs)

        self.means = tf.keras.layers.Dense(original_dim, activation='linear')
        self.var = tf.keras.layers.Dense(original_dim, activation='linear')

    def call(self, inputs, **kwargs):
        activation, sf = inputs
        mean, var = self.means(activation), self.var(activation)

        # clip to log of largest values supported by log operation
        bound = 60.
        mean_clip = tf.clip_by_value(mean, -bound, bound, "decoder_clip")
        var_clip = tf.clip_by_value(var, -bound, bound, "decoder_clip")

        invlinker_mean = tf.exp(mean_clip + sf)
        invlinker_var = tf.exp(var_clip)

        return [invlinker_mean, invlinker_var]


class NegBinSharedDispOutput(tf.keras.layers.Layer):
    """Negative binomial output layer with a single dispersion estimate per features"""

    def __init__(
            self,
            original_dim=None,
            name='neg_bin_shared_disp_output',
            **kwargs
    ):

        super().__init__(name=name, **kwargs)

        self.means = tf.keras.layers.Dense(original_dim, activation='linear')
        self.var = self.add_weight(
            "var_bias",
            shape=[1, original_dim]
        )

    def call(self, inputs, **kwargs):
        activation, sf = inputs
        mean = self.means(activation)
        var = self.var
        var = tf.broadcast_to(var, tf.shape(mean))

        # clip to log of largest values supported by log operation
        bound = 60.
        mean_clip = tf.clip_by_value(mean, -bound, bound, "decoder_clip")
        var_clip = tf.clip_by_value(var, -bound, bound, "decoder_clip")

        invlinker_mean = tf.exp(mean_clip + sf)
        invlinker_var = tf.exp(var_clip)

        return [invlinker_mean, invlinker_var]


class NegBinConstDispOutput(tf.keras.layers.Layer):
    """Negative binomial output layer with dispersion set as constant (=1)."""

    def __init__(
            self,
            original_dim=None,
            name='neg_bin_const_disp_output',
            **kwargs
    ):

        super().__init__(name=name, **kwargs)

        self.means = tf.keras.layers.Dense(original_dim, activation='linear')
        self.var_constant = 1.

    def call(self, inputs, **kwargs):
        activation, sf = inputs
        mean = self.means(activation)
        var = tf.constant([[self.var_constant]], dtype=activation.dtype)
        var = tf.broadcast_to(var, tf.shape(mean))

        # clip to log of largest values supported by log operation
        bound = 60.
        mean_clip = tf.clip_by_value(mean, -bound, bound, "decoder_clip")
        var_clip = tf.clip_by_value(var, -bound, bound, "decoder_clip")

        invlinker_mean = tf.exp(mean_clip + sf)
        invlinker_var = tf.exp(var_clip)

        return [invlinker_mean, invlinker_var]


class GaussianOutput(tf.keras.layers.Layer):
    """
    Gaussian output layer.

    Size factor only makes sense if logged and data is positive and logged.
    """

    def __init__(
            self,
            original_dim=None,
            name='gaussian_output',
            **kwargs
    ):

        super().__init__(name=name, **kwargs)

        self.means = tf.keras.layers.Dense(original_dim, activation='linear')
        self.var = tf.keras.layers.Dense(original_dim, activation='linear')

    def call(self, inputs, **kwargs):
        activation, sf = inputs
        mean, var = self.means(activation), self.var(activation)

        # clip to log of largest values supported by log operation
        bound = 60.
        mean_clip = tf.clip_by_value(mean, tf.exp(-bound), tf.exp(bound), "decoder_clip")
        var_clip = tf.clip_by_value(var, -bound, bound, "decoder_clip")

        invlinker_mean = mean_clip + sf
        invlinker_var = tf.exp(var_clip)

        return [invlinker_mean, invlinker_var]


class GaussianSharedStdOutput(tf.keras.layers.Layer):
    """
    Gaussian output layer with a single standard deviation estimate per features.

    Size factor only makes sense if logged and data is positive and logged.
    """

    def __init__(
            self,
            original_dim=None,
            name='gaussian_shared_disp_output',
            **kwargs
    ):

        super().__init__(name=name, **kwargs)

        self.means = tf.keras.layers.Dense(original_dim, activation='linear')
        self.var = self.add_weight(
            "var_bias",
            shape=[1, original_dim]
        )

    def call(self, inputs, **kwargs):
        activation, sf = inputs
        mean = self.means(activation)
        var = self.var
        var = tf.broadcast_to(var, tf.shape(mean))

        # clip to log of largest values supported by log operation
        bound = 60.
        mean_clip = tf.clip_by_value(mean, tf.exp(-bound), tf.exp(bound), "decoder_clip")
        var_clip = tf.clip_by_value(var, -bound, bound, "decoder_clip")

        invlinker_mean = mean_clip + sf
        invlinker_var = tf.exp(var_clip)

        return [invlinker_mean, invlinker_var]


class GaussianConstStdOutput(tf.keras.layers.Layer):
    """
    Gaussian output layer with standard deviation set as constant (=1).

    Size factor only makes sense if logged and data is positive and logged.
    """

    def __init__(
            self,
            original_dim=None,
            name='gaussian_const_disp_output',
            **kwargs
    ):

        super().__init__(name=name, **kwargs)

        self.means = tf.keras.layers.Dense(original_dim, activation='linear')
        self.var_constant = 1.

    def call(self, inputs, **kwargs):
        activation, sf = inputs
        mean = self.means(activation)
        var = tf.constant([[self.var_constant]], dtype=activation.dtype)
        var = tf.broadcast_to(var, tf.shape(mean))

        # clip to log of largest values supported by log operation
        bound = 60.
        mean_clip = tf.clip_by_value(mean, tf.exp(-bound), tf.exp(bound), "decoder_clip")
        var_clip = tf.clip_by_value(var, -bound, bound, "decoder_clip")

        invlinker_mean = mean_clip + sf
        invlinker_var = tf.exp(var_clip)

        return [invlinker_mean, invlinker_var]



import os
import warnings
from typing import List, Union

import anndata
import numpy as np
import pandas as pd

from sfaira.consts import AdataIdsSfaira, OCS, AdataIds
from sfaira.data.store.stores.base import StoreBase
from sfaira.data.store.stores.multi import StoresAnndata
from sfaira.data.store.stores.single import StoreSingleFeatureSpace
from sfaira.models.base import BasicModel
from sfaira.versions.metadata import CelltypeUniverse, OntologyCl, OntologyObo
from sfaira.versions.topologies import TopologyContainer


def split_idx(data: StoreSingleFeatureSpace, test_split, val_split):
    """
    Split training and evaluation data.
    """
    np.random.seed(1)
    all_idx = np.arange(0, data.n_obs)  # n_obs is both a property of AnnData and DistributedStoreBase
    if isinstance(test_split, float) or isinstance(test_split, int):
        idx_test = np.sort(np.random.choice(
            a=all_idx,
            size=round(data.n_obs * test_split),
            replace=False,
        ))
    elif isinstance(test_split, dict):
        in_test = np.ones((data.n_obs,), dtype=int) == 1
        for k, v in test_split.items():
            if isinstance(v, bool) or isinstance(v, int) or isinstance(v, str):
                v = [v]
            elif isinstance(v, tuple):
                v = list(v)
            elif isinstance(v, np.ndarray):
                v = v.tolist()
            if not isinstance(v, list):
                raise ValueError(f"conversion of v {v} to list failed")
            if np.any([isinstance(vi, list) or isinstance(vi, tuple) or isinstance(vi, np.ndarray) for vi in v]):
                raise ValueError(f"found nested list v {v}, only use bool, scalar or string elements in v.")
            idx = data.get_subset_idx(attr_key=k, values=v, excluded_values=None)
            # Build continuous vector across all sliced data sets and establish which observations are kept
            # in subset.
            in_test_k = np.ones((data.n_obs,), dtype=int) == 0
            counter = 0
            for kk, vv in data.indices.items():
                if kk in idx.keys() and len(idx[kk]) > 0:
                    in_test_k[np.where([x in idx[kk] for x in vv])[0] + counter] = True
                counter += len(vv)
            in_test = np.logical_and(in_test, in_test_k)
        idx_test = np.sort(np.where(in_test)[0])
    else:
        raise ValueError("type of test_split %s not recognized" % type(test_split))
    print(f"Found {len(idx_test)} out of {data.n_obs} cells that correspond to test data set")
    assert len(idx_test) < data.n_obs, f"test set covers full data set, apply a more restrictive test " \
                                       f"data definiton ({len(idx_test)}, {data.n_obs})"
    idx_train_eval = all_idx[~np.isin(all_idx, idx_test)]
    np.random.seed(1)
    idx_eval = np.sort(np.random.choice(
        a=idx_train_eval,
        size=round(len(idx_train_eval) * val_split),
        replace=False
    ))
    idx_train = np.sort(idx_train_eval[~np.isin(idx_train_eval, idx_eval)])

    # Check that none of the train, test, eval partitions are empty
    if not len(idx_test):
        warnings.warn("Test partition is empty!")
    if not len(idx_eval):
        raise ValueError("The evaluation dataset is empty.")
    if not len(idx_train):
        raise ValueError("The train dataset is empty.")
    return idx_train, idx_eval, idx_test


class EstimatorBase:

    data: StoreSingleFeatureSpace
    model: Union[BasicModel, None]
    topology_container: TopologyContainer
    model_id: Union[str, None]
    weights: Union[np.ndarray, None]
    model_dir: Union[str, None]
    history: Union[dict, None]
    train_hyperparam: Union[dict, None]
    idx_train: Union[np.ndarray, None]
    idx_eval: Union[np.ndarray, None]
    idx_test: Union[np.ndarray, None]
    adata_ids: AdataIds

    def __init__(
            self,
            data: Union[anndata.AnnData, List[anndata.AnnData], StoreSingleFeatureSpace],
            model_dir: Union[str, None],
            model_class: str,
            model_id: Union[str, None],
            model_topology: TopologyContainer,
            weights_md5: Union[str, None] = None,
            cache_path: str = os.path.join('cache', ''),
            adata_ids: AdataIds = AdataIdsSfaira()
    ):
        self.model = None
        self.model_dir = model_dir
        self.model_id = model_id
        self.model_class = model_class
        self.topology_container = model_topology
        if isinstance(data, anndata.AnnData):
            data = StoresAnndata(adatas=data).stores[self.organism]
        if isinstance(data, list) or isinstance(data, tuple):
            for x in data:
                assert isinstance(x, anndata.AnnData), f"found element in list that was not anndata but {type(x)}"
            data = StoresAnndata(adatas=data).stores[self.organism]
        self.data = data
        # Prepare store with genome container sub-setting:
        # This class is tailored for DistributedStoreSingleFeatureSpace but we test for the base class here in the
        # constructor so that genome_container can also be set in inheriting classes that may be centred around
        # different child classes of DistributedStoreBase.
        if isinstance(self.data, StoreBase):
            self.data.genome_container = self.topology_container.gc

        self.history = None
        self.train_hyperparam = None
        self.idx_train = None
        self.idx_eval = None
        self.idx_test = None
        self.md5 = weights_md5
        self.cache_path = cache_path
        self._adata_ids = adata_ids

    @property
    def model_type(self):
        return self.topology_container.model_type

    @property
    def organism(self):
        return self.topology_container.organism

    def _get_class_dict(
            self,
            obs_key: str
    ):
        y = self.obs[obs_key]
        for i, val in enumerate(y):
            if type(val) == list:
                y[i] = " / ".join(val)
        labels = np.unique(y)
        label_dict = {}
        for i, label in enumerate(labels):
            label_dict.update({label: float(i)})
        return label_dict

    def split_train_val_test(self, val_split: float, test_split: Union[float, dict]):
        """
        Split indices in store into train, valiation and test split.
        """
        idx_train, idx_eval, idx_test = split_idx(data=self.data, test_split=test_split, val_split=val_split)
        self.idx_train = idx_train
        self.idx_eval = idx_eval
        self.idx_test = idx_test

    @property
    def using_store(self) -> bool:
        return isinstance(self.data, StoreSingleFeatureSpace)

    @property
    def obs_train(self) -> pd.DataFrame:
        return self.data.checkout(idx=self.idx_train).obs

    @property
    def obs_eval(self) -> pd.DataFrame:
        return self.data.checkout(idx=self.idx_eval).obs

    @property
    def obs_test(self) -> pd.DataFrame:
        return self.data.checkout(idx=self.idx_test).obs

    @property
    def obs(self) -> pd.DataFrame:
        return self.data.checkout(idx=self._process_idx_for_eval(None)).obs

    def _process_idx_for_eval(self, idx):
        """
        Defaults to all observations if no indices are defined.
        """
        if idx is None:
            idx = np.arange(0, self.data.n_obs)
        return idx


class EstimatorBaseEmbedding(EstimatorBase):

    pass


class EstimatorBaseCelltype(EstimatorBase):

    def __init__(
            self,
            data: Union[anndata.AnnData, StoreSingleFeatureSpace],
            model_dir: Union[str, None],
            model_id: Union[str, None],
            model_topology: TopologyContainer,
            weights_md5: Union[str, None] = None,
            cache_path: str = os.path.join('cache', ''),
            celltype_ontology: Union[OntologyObo, None] = None,
            max_class_weight: float = 1e3,
            remove_unlabeled_cells: bool = True,
            adata_ids: AdataIds = AdataIdsSfaira()
    ):
        super(EstimatorBaseCelltype, self).__init__(
            data=data,
            model_dir=model_dir,
            model_class="celltype",
            model_id=model_id,
            model_topology=model_topology,
            weights_md5=weights_md5,
            cache_path=cache_path,
            adata_ids=adata_ids
        )

        if remove_unlabeled_cells:
            # Remove cells without type label from store:
            self.data.subset(attr_key="cell_type", excluded_values=[
                self._adata_ids.unknown_metadata_identifier,
                self._adata_ids.not_a_cell_celltype_identifier,
            ])
        assert "cl" in self.topology_container.output.keys(), self.topology_container.output.keys()
        assert "targets" in self.topology_container.output.keys(), self.topology_container.output.keys()
        self.max_class_weight = max_class_weight
        if celltype_ontology is None:
            celltype_ontology = OntologyCl(branch=self.topology_container.output["cl"])
        self.celltype_universe = CelltypeUniverse(
            cl=celltype_ontology,
            uberon=OCS.organ,
        )
        # Set leaves if they are defined in topology:
        if self.topology_container.output["targets"] is not None:
            self.celltype_universe.onto_cl.leaves = self.topology_container.output["targets"]

    @property
    def ntypes(self):
        return self.celltype_universe.onto_cl.n_leaves

    @property
    def ontology_ids(self):
        return self.celltype_universe.onto_cl.convert_to_id(self.celltype_universe.onto_cl.leaves)

    @property
    def ontology_names(self):
        return self.celltype_universe.onto_cl.convert_to_name(self.celltype_universe.onto_cl.leaves)

    def _one_hot_encoder(self):
        leave_maps = self.celltype_universe.onto_cl.prepare_maps_to_leaves(include_self=True)

        def encoder(x) -> np.ndarray:
            if isinstance(x, str):
                x = [x]
            # Encodes unknowns to empty rows.
            idx = [
                leave_maps[y] if y not in [
                    self._adata_ids.unknown_metadata_identifier,
                    self._adata_ids.not_a_cell_celltype_identifier,
                ] else np.array([])
                for y in x
            ]
            oh = np.zeros((len(x), self.ntypes,), dtype="float32")
            for i, y in enumerate(idx):
                scale = len(y)
                if scale > 0:
                    oh[i, y] = 1. / scale
            return oh

        return encoder

    def _get_celltype_out(self, idx: Union[np.ndarray, None]):
        """
        TODO depreceate, carry over weight code to _get_generator
        Build one hot encoded cell type output tensor and observation-wise weight matrix.
        """
        if idx is None:
            idx = np.arange(0, self.data.n_obs)
        # One whether "unknown" is already included, otherwise add one extra column.
        onehot_encoder = self._one_hot_encoder()
        y = np.concatenate([
            onehot_encoder(z)
            for z in self.obs[self._adata_ids.cell_type + self._adata_ids.onto_id_suffix].values[idx].tolist()
        ], axis=0)
        # Distribute aggregated class weight for computation of weights:
        freq = np.mean(y / np.sum(y, axis=1, keepdims=True), axis=0, keepdims=True)
        weights = 1. / np.matmul(y, freq.T)  # observation wise weight matrix
        # Threshold weights:
        weights = np.asarray(
            np.minimum(weights, np.zeros_like(weights) + self.max_class_weight),
            dtype="float32"
        ).flatten()
        return weights, y



from sfaira.consts.adata_fields import AdataIds, AdataIdsSfaira, AdataIdsCellxgene, AdataIdsCellxgene_v2_0_0
from sfaira.consts.meta_data_files import META_DATA_FIELDS
from sfaira.consts.ontologies import OntologyContainerSfaira, OTHER_ORGANISM_KEY
from sfaira.consts.utils import clean_cache

OCS = OntologyContainerSfaira()



from __future__ import annotations

import anndata
import multiprocessing
import numpy as np
import pandas as pd
import os
from os import PathLike
import pandas
import pydoc
import sys
from typing import Dict, List, Union
import warnings

from sfaira.consts import AdataIds, AdataIdsSfaira
from sfaira.data.dataloaders.base.dataset import DatasetBase
from sfaira.data.dataloaders.base.utils import identify_tsv, is_child
from sfaira.versions.genomes.genomes import GenomeContainer
from sfaira.versions.metadata.maps import prepare_ontology_map_tab
from sfaira.data.utils import read_yaml

UNS_STRING_META_IN_OBS = "__obs__"


def map_fn(inputs):
    """
    Functional to load data set with predefined additional actions.

    :param inputs:
    :return: None if function ran, error report otherwise
    """
    ds, load_raw, allow_caching, kwargs, func, kwargs_func = \
        inputs
    try:
        ds.load(load_raw=load_raw, allow_caching=allow_caching, **kwargs)
        if func is not None:
            x = func(ds, **kwargs_func)
            ds.clear()
            return x
        else:
            return None
    except (FileNotFoundError, OSError) as e:
        return ds.id, e,


def merge_uns_from_list(adata_ls):
    """
    Merge .uns from list of adata objects.

    Merges values for innert join of keys across all objects. This will retain uns streamlining.
    Keeps shared uns values for a given key across data sets as single value (not a list of 1 unique value).
    Other values are represented as a list of all unique values found.
    """
    uns_keys = [list(x.uns.keys()) for x in adata_ls]
    uns_keys_shared = set(uns_keys[0])
    for x in uns_keys[1:]:
        uns_keys_shared = uns_keys_shared.intersection(set(x))
    uns_keys_shared = list(uns_keys_shared)
    uns = {}
    for k in uns_keys_shared:
        uns_k = []
        for y in adata_ls:
            x = y.uns[k]
            if isinstance(x, list):
                pass
            elif isinstance(x, tuple):
                x = list(x)
            elif isinstance(x, np.ndarray):
                x = x.tolist()
            else:
                x = [x]
            uns_k.extend(x)
        uns_k = np.unique(uns_k).tolist()
        if len(uns_k) == 1:
            uns_k = uns_k[0]
        uns[k] = uns_k
    return uns


load_doc = \
    """
    :param remove_gene_version: Remove gene version string from ENSEMBL ID so that different versions in different data sets are superimposed.
    :param match_to_reference: Reference genomes name or False to keep original feature space.
    :param load_raw: Loads unprocessed version of data if available in data loader.
    :param allow_caching: Whether to allow method to cache adata object for faster re-loading.
    """


class DatasetGroup:
    """
    Container class that co-manages multiple data sets, removing need to call Dataset() methods directly through
    wrapping them.

    Example:

    #query loaders lung
    #from sfaira.dev.data.loaders.lung import DatasetGroupLung as DatasetGroup
    #dsg_humanlung = DatasetGroupHuman(path='path/to/data')
    #dsg_humanlung.load_all(match_to_reference='Homo_sapiens_GRCh38_97')
    #dsg_humanlung[some_id]
    #dsg_humanlung.adata
    """
    datasets: Dict[str, DatasetBase]
    _collection_id: str

    def __init__(self, datasets: dict, collection_id: str = "default"):
        self._adata_ids = AdataIdsSfaira()
        self.datasets = datasets
        self._collection_id = collection_id

    def load(
            self,
            annotated_only: bool = False,
            load_raw: bool = False,
            allow_caching: bool = True,
            processes: int = 1,
            func=None,
            kwargs_func: Union[None, dict] = None,
            verbose: int = 0,
            **kwargs
    ):
        """
        Load all datasets in group (option for temporary loading).

        Note: This method automatically subsets to the group to the data sets for which input files were found.

        This method also allows temporarily loading data sets to execute function on loaded data sets (supply func).
        In this setting, datasets are removed from memory after the function has been executed.

        :param annotated_only:
        :param load_raw:
        :param allow_caching:
        :param processes: Processes to parallelise loading over. Uses python multiprocessing if > 1, for loop otherwise.
        :param func: Function to run on loaded datasets. map_fun should only take one argument, which is a Dataset
            instance. The return can be empty:

                def func(dataset, **kwargs_func):
                    # code manipulating dataset and generating output x.
                    return x
        :param kwargs_func: Kwargs of func.
        :param verbose: Verbosity of description of loading failure.

                - 0: no indication of failure
                - 1: indication of which data set failed in warning
                - 2: 1 with error report in warning
                - 3: reportin as in 2 but aborts with OSError
        """
        args = [
            load_raw,
            allow_caching,
            kwargs,
            func,
            kwargs_func
        ]

        if processes > 1 and len(self.datasets.items()) > 1:  # multiprocessing parallelisation
            print(f"using python multiprocessing (processes={processes}), "
                  f"for easier debugging revert to sequential execution (processes=1)")
            with multiprocessing.Pool(processes=processes) as pool:
                res = pool.starmap(map_fn, [
                    (tuple([v] + args),)
                    for k, v in self.datasets.items() if v.annotated or not annotated_only
                ])
            # Clear data sets that were not successfully loaded because of missing data:
            for x in res:
                if x is not None:
                    print(x[1])
                    del self.datasets[x[0]]
        else:  # for loop
            datasets_to_remove = []
            for k, v in self.datasets.items():
                print(f"loading {k}")
                x = map_fn(tuple([v] + args))
                # Clear data sets that were not successfully loaded because of missing data:
                if x is not None:
                    if verbose == 1:
                        warnings.warn(f"data set {k} not loaded")
                    if verbose == 2:
                        warnings.warn(f"data set {k} not loaded\nin data set {x[0]}: {x[1]}")
                    if verbose == 3:
                        raise OSError(f"data set {k} not loaded\nin data set {x[0]}: {x[1]}")
                    datasets_to_remove.append(k)
            for k in datasets_to_remove:
                del self.datasets[k]

    load.__doc__ += load_doc

    def streamline_metadata(
            self,
            schema: str = "sfaira",
            clean_obs: bool = True,
            clean_var: bool = True,
            clean_uns: bool = True,
            clean_obs_names: bool = True,
            keep_orginal_obs: bool = False,
            keep_symbol_obs: bool = True,
            keep_id_obs: bool = True,
    ):
        """
        Streamline the adata instance in each data set to output format.
        Output format are saved in ADATA_FIELDS* classes.

        :param schema: Export format.
            - "sfaira"
            - "cellxgene"
        :param clean_obs: Whether to delete non-streamlined fields in .obs, .obsm and .obsp.
        :param clean_var: Whether to delete non-streamlined fields in .var, .varm and .varp.
        :param clean_uns: Whether to delete non-streamlined fields in .uns.
        :param clean_obs_names: Whether to replace obs_names with a string comprised of dataset id and an increasing integer.
        :param clean_obs_names: Whether to replace obs_names with a string comprised of dataset id and an increasing
            integer.
        :param keep_orginal_obs: For ontology-constrained .obs columns, whether to keep a column with original
            annotation.
        :param keep_symbol_obs: For ontology-constrained .obs columns, whether to keep a column with ontology symbol
            annotation.
        :param keep_id_obs: For ontology-constrained .obs columns, whether to keep a column with ontology ID annotation.
        :return:
        """
        for x in self.ids:
            self.datasets[x].streamline_metadata(
                schema=schema,
                clean_obs=clean_obs,
                clean_var=clean_var,
                clean_uns=clean_uns,
                clean_obs_names=clean_obs_names,
                keep_orginal_obs=keep_orginal_obs,
                keep_symbol_obs=keep_symbol_obs,
                keep_id_obs=keep_id_obs,
            )

    def streamline_features(
            self,
            match_to_release: Union[str, Dict[str, str], None] = None,
            remove_gene_version: bool = True,
            subset_genes_to_type: Union[None, str, List[str]] = None,
            schema: Union[str, None] = None,
    ):
        """
        Subset and sort genes to genes defined in an assembly or genes of a particular type, such as protein coding.
        :param match_to_release: Which genome annotation release to map the feature space to. Note that assemblies from
            ensbeml are usually named as Organism.Assembly.Release, this is the Release string. Can be:
                - str: Provide the name of the release.
                - dict: Mapping of organism to name of the release (see str format). Chooses release for each
                    data set based on organism annotation.:param remove_gene_version: Whether to remove the version
                    number after the colon sometimes found in ensembl gene ids.
        :param subset_genes_to_type: Type(s) to subset to. Can be a single type or a list of types or None.
            Types can be:
                - None: All genes in assembly.
                - "protein_coding": All protein coding genes in assembly.
        """
        for x in self.ids:
            self.datasets[x].streamline_features(
                match_to_release=match_to_release,
                remove_gene_version=remove_gene_version,
                subset_genes_to_type=subset_genes_to_type,
                schema=schema,
            )

    def collapse_counts(self):
        """
        Collapse count matrix along duplicated index.
        """
        for x in self.ids:
            self.datasets[x].collapse_counts()

    def write_distributed_store(
            self,
            dir_cache: Union[str, os.PathLike],
            store_format: str = "dao",
            dense: bool = False,
            compression_kwargs: dict = {},
            chunks: Union[int, None] = None,
    ):
        """
        Write data set into a format that allows distributed access to data set on disk.

        Stores are useful for distributed access to data sets, in many settings this requires some streamlining of the
        data sets that are accessed. Use .streamline_* before calling this method to streamline the data sets.
        This method writes a separate file for each data set in this object.

        :param dir_cache: Directory to write cache in.
        :param store_format: Disk format for objects in cache. Recommended is "dao".

            - "h5ad": Allows access via backed .h5ad.
                Note on compression: .h5ad supports sparse data with is a good compression that gives fast row-wise
                    access if the files are csr, so further compression potentially not necessary.
            - "dao": Distributed access optimised format, recommended for batched access in optimisation, for example.
        :param dense: Whether to write sparse or dense store, this will be homogenously enforced.
        :param compression_kwargs: Compression key word arguments to give to h5py or zarr
            For store_format=="h5ad", see also anndata.AnnData.write_h5ad:
                - compression,
                - compression_opts.
            For store_format=="dao", see also sfaira.data.write_dao which relays kwargs to
            zarr.hierarchy.create_dataset:
                - compressor
                - overwrite
                - order
                and others.
        :param chunks: Observation axes of chunk size of zarr array, see anndata.AnnData.write_zarr documentation.
            Only relevant for store=="dao". The feature dimension of the chunks is always is the full feature space.
            Uses zarr default chunking across both axes if None.
        """
        for _, v in self.datasets.items():
            v.write_distributed_store(dir_cache=dir_cache, store_format=store_format, dense=dense,
                                      compression_kwargs=compression_kwargs, chunks=chunks)

    def write_backed(
            self,
            adata_backed: anndata.AnnData,
            genome: str,
            idx: List[np.ndarray],
            annotated_only: bool = False,
            load_raw: bool = False,
            allow_caching: bool = True,
    ):
        """
        Loads data set group into slice of backed anndata object.

        Subsets self.datasets to the data sets that were found. Note that feature space is automatically formatted as
        this is necessary for concatenation.

        :param adata_backed: Anndata instance to load into.
        :param genome: Genome container target genomes loaded.
        :param idx: Indices in adata_backed to write observations to. This can be used to immediately create a
            shuffled object. This has to be a list of the length of self.data, one index array for each dataset.
        :param annotated_only:
        :param load_raw: See .load().
        :param allow_caching: See .load().
        :return: New row index for next element to be written into backed anndata.
        """
        i = 0
        for x in self.ids:
            # if this is for celltype prediction, only load the data with have celltype annotation
            try:
                if self.datasets[x].annotated or not annotated_only:
                    self.datasets[x].write_backed(
                        adata_backed=adata_backed,
                        genome=genome,
                        idx=idx[i],
                        load_raw=load_raw,
                        allow_caching=allow_caching
                    )
                    i += 1
            except FileNotFoundError:
                del self.datasets[x]

    def write_ontology_class_maps(
            self,
            fn,
            attrs: List[str],
            protected_writing: bool = True,
            **kwargs
    ):
        """
        Write cell type maps of free text cell types to ontology classes.

        :param fn: File name of tsv to write class maps to.
        :param attrs: Attributes to create a tsv for. Must correspond to *_obs_key in yaml.
        :param protected_writing: Only write if file was not already found.
        """
        for x in attrs:
            # TODO need to deal with groups that contain multiple organisms here.
            organism = np.unique([v.organism for v in self.datasets.values()])
            if len(organism) == 1:
                organism = organism[0]
            else:
                raise ValueError(f"write_ontology_class_maps() for mixed organisms not yet supported {organism}.")
            fn_x = fn + "_" + x + ".tsv"
            labels_original = []
            organs = []
            for v in self.datasets.values():
                obs_key = getattr(v, x + "_obs_key")
                if obs_key is not None and obs_key in v.adata.obs.columns:
                    labels_original.append(np.unique(v.adata.obs[obs_key].values))
                if v.organ is not None:
                    organs.append(v.organ)
            if len(labels_original) == 0:
                warnings.warn(f"Attempted to write ontology class-maps for meta data {x} without corresponding "
                              "annotation in any data set.")
            else:
                labels_original = np.sort(np.unique(np.concatenate(labels_original)))
                # Only use anatomic constraint if all data sets are from same organ.
                # TODO this could be extended in the future.
                organs = np.unique(organs)
                organs = organs[0] if len(organs) == 1 else None
                tab = prepare_ontology_map_tab(
                    source=labels_original,
                    onto=x,
                    organism=organism,
                    anatomical_constraint=organs,
                    choices_for_perfect_match=False,
                    include_synonyms=True,
                    match_only=False,
                    omit_list=[v._adata_ids.not_a_cell_celltype_identifier,
                               v._adata_ids.unknown_metadata_identifier],
                    **kwargs
                )
                tab = tab.sort_values(self._adata_ids.classmap_source_key)
                if not os.path.exists(fn_x) or not protected_writing:
                    tab.to_csv(fn_x, index=False, sep="\t")

    def download(self, **kwargs):
        for _, v in self.datasets.items():
            v.download(**kwargs)

    @property
    def ids(self):
        return list(self.datasets.keys())

    @property
    def collection_id(self):
        return self._collection_id

    @collection_id.setter
    def collection_id(self, x: str):
        self._collection_id = x

    @property
    def adata_ls(self):
        adata_ls = []
        for i in self.ids:
            if self.datasets[i] is not None:
                if self.datasets[i].adata is not None:
                    adata_ls.append(self.datasets[i].adata)
        return adata_ls

    @property
    def adata(self):
        adata_ls = self.adata_ls
        if not adata_ls:
            return None
        if len(adata_ls) == 1:
            for i in self.ids:
                if self.datasets[i] is not None:
                    if self.datasets[i].adata is not None:
                        ds_id = i
            adata_concat = adata_ls[0]
            adata_concat.obs[self._adata_ids.dataset] = ds_id
        else:
            # Check that all individual adata objects in linked Dataset instances have identicall streamlined features
            # and metadata
            match_ref_list = []
            rm_gene_ver_list = []
            gene_type_list = []
            for d_id in self.ids:
                if self.datasets[d_id].adata is not None:
                    assert self.datasets[d_id].mapped_features, \
                        f"Dataset {d_id} does not seem to have a streamlined " \
                        f"featurespace. To obtain an adata object from this " \
                        f"DatasetGroup, all contained Datasets need to have a " \
                        f"streamlined featurespace. Run .streamline_features()" \
                        f" first."
                    assert self.datasets[d_id].streamlined_meta, \
                        f"Dataset {d_id} does not seem to have streamlined " \
                        f"metadata. To obtain an adata object from this " \
                        f"DatasetGroup, all contained Datasets need to have " \
                        f"streamlined metadata. Run .streamline_metadata() first."
                    match_ref_list.append(self.datasets[d_id].mapped_features)
                    rm_gene_ver_list.append(self.datasets[d_id].remove_gene_version)
                    gene_type_list.append(self.datasets[d_id].subset_gene_type)
            assert len(set(match_ref_list)) == 1, \
                "Not all datasets in this group had their features matched to the same reference (argument " \
                "'match_to_reference' of method .streamline_features())." \
                "This is however a prerequisite for creating a combined adata object."
            assert len(set(rm_gene_ver_list)) == 1, \
                "Not all datasets in this group have had their gene version removed (argument 'remove_gene_version' " \
                "of method .streamline_features()). This is however a prerequisite for creating a combined adata " \
                "object."
            assert len(set(gene_type_list)) == 1, \
                "Not all datasets in this group had their featurespace subsetted to the same gene type (argument " \
                "'subset_gene_type' of method .streamline_features()). This is however a prerequisite for creating a " \
                "combined adata object."

            var_original = adata_ls[0].var.copy()
            for a in adata_ls:
                a.var_names_make_unique()
            # TODO: need to keep this? -> yes, still catching errors here (March 2020)
            # Fix for loading bug: sometime concatenating sparse matrices fails the first time but works on second try.
            try:
                adata_concat = adata_ls[0].concatenate(
                    *adata_ls[1:],
                    join="outer",
                    batch_key=self._adata_ids.dataset,
                    batch_categories=[i for i in self.ids if self.datasets[i].adata is not None],
                    index_unique=None
                )
            except ValueError:
                adata_concat = adata_ls[0].concatenate(
                    *adata_ls[1:],
                    join="outer",
                    batch_key=self._adata_ids.dataset,
                    batch_categories=[i for i in self.ids if self.datasets[i].adata is not None],
                    index_unique=None
                )
            adata_concat.var = var_original
            adata_concat.uns = merge_uns_from_list(adata_ls)
            adata_concat.uns[self._adata_ids.mapped_features] = match_ref_list[0]

        return adata_concat

    def obs_concat(self, keys: Union[list, None] = None):
        """
        Returns concatenation of all .obs.

        Uses union of all keys if keys is not provided.

        :param keys:
        :return:
        """
        if keys is None:
            keys = np.unique(np.concatenate([list(x.obs.columns) for x in self.adata_ls]))
        obs_concat = pandas.concat([pandas.DataFrame(dict(
            [
                (k, self.datasets[x].adata.obs[k]) if k in self.datasets[x].adata.obs.columns
                else (k, ["nan" for _ in range(self.datasets[x].adata.obs.shape[0])])
                for k in keys
            ] + [(self._adata_ids.dataset, [x for _ in range(self.datasets[x].adata.obs.shape[0])])]
        )) for x in self.ids if self.datasets[x].adata is not None])
        return obs_concat

    def ncells_bydataset(self, annotated_only: bool = False) -> np.ndarray:
        cells = []
        for x in self.ids:
            # if this is for celltype prediction, only load the data with have celltype annotation
            try:
                if self.datasets[x].annotated or not annotated_only:
                    cells.append(self.datasets[x].ncells)
            except FileNotFoundError:
                del self.datasets[x]
        return np.asarray(cells)

    def ncells(self, annotated_only: bool = False):
        cells = self.ncells_bydataset(annotated_only=annotated_only)
        return np.sum(cells)

    @property
    def ontology_container_sfaira(self):
        return self.datasets[self.ids[0]].ontology_container_sfaira

    @property
    def ontology_celltypes(self):
        """
        # TODO: use might be replaced by ontology_container_sfaira in the future.
        """
        print("deprecation warning for ontology_celltypes()")
        return self.datasets[self.ids[0]].ontology_container_sfaira.cell_type

    def project_celltypes_to_ontology(self, adata_fields: Union[AdataIds, None] = None, copy=False):
        """
        Project free text cell type names to ontology based on mapping table.
        :return:
        """
        for _, v in self.datasets.items():
            v.project_free_to_ontology(adata_fields=adata_fields, copy=copy)

    def subset(self, key, values: Union[list, tuple, np.ndarray]):
        """
        Subset list of adata objects based on sample-wise properties.

        These keys are properties that are available in lazy model.
        Subsetting happens on .datasets.

        :param key: Property to subset by.
        :param values: Classes to overlap to. Return if elements match any of these classes.
        :return:
        """
        ids_del = []
        if isinstance(values, np.ndarray):
            values = values.tolist()
        if isinstance(values, tuple):
            values = list(values)
        if not isinstance(values, list):
            values = [values]
        for x in self.ids:
            try:
                values_found = getattr(self.datasets[x], key)
            except AttributeError:
                raise ValueError(f"{key} not a valid property of data set object")
            try:
                ontology = getattr(self.datasets[x].ontology_container_sfaira, key)
            except AttributeError:
                raise ValueError(f"{key} not a valid property of ontology_container object")
            if values_found is None:
                # Delete entries which do not have this meta data item annotated.
                ids_del.append(x)
            else:
                if not isinstance(values_found, list):
                    values_found = [values_found]
                if not np.any([
                    np.any([
                        is_child(query=y, ontology=ontology, ontology_parent=z)
                        for z in values
                    ]) for y in values_found
                ]):
                    # Delete entries which a non-matching meta data value associated with this item.
                    ids_del.append(x)
        for x in ids_del:
            del self.datasets[x]

    def subset_cells(self, key, values: Union[str, List[str]]):
        """
        Subset list of adata objects based on cell-wise properties.

        These keys are properties that are not available in lazy model and require loading first because the
        subsetting works on the cell-level: .adata are maintained but reduced to matches.

        :param key: Property to subset by. Options:

            - "assay_differentiation" points to self.assay_differentiation_obs_key
            - "assay_sc" points to self.assay_sc_obs_key
            - "assay_type_differentiation" points to self.assay_type_differentiation_obs_key
            - "cell_line" points to self.cell_line
            - "cell_type" points to self.cell_type_obs_key
            - "developmental_stage" points to self.developmental_stage_obs_key
            - "ethnicity" points to self.ethnicity_obs_key
            - "organ" points to self.organ_obs_key
            - "organism" points to self.organism_obs_key
            - "sample_source" points to self.sample_source_obs_key
            - "sex" points to self.sex_obs_key
            - "state_exact" points to self.state_exact_obs_key
        :param values: Classes to overlap to.
        :return:
        """
        for x in self.ids:
            self.datasets[x].subset_cells(key=key, values=values)
            if self.datasets[x].adata is None:  # No observations (cells) left.
                del self.datasets[x]

    @property
    def additional_annotation_key(self) -> Dict[str, Union[None, str]]:
        """"
        Return dictionary of additional_annotation_key for each data set with ids as keys.
        """
        return dict([
            (k, self.datasets[k].additional_annotation_key)
            for k, v in self.datasets.items()
        ])

    @additional_annotation_key.setter
    def additional_annotation_key(self, x: Dict[str, Union[None, str]]):
        """
        Allows setting of additional_annotation_key in a subset of datasets identifed by keys in x.

        :param x: Dictionary with data set ids in keys and new _additional_annotation_key values to be setted in values.
            Note that you can either add  or change secondary annotation by setting a value to a string or remove it
            by setting a value to None.
        :return:
        """
        for k, v in x.items():
            self.datasets[k].additional_annotation_key = v

    @property
    def doi(self) -> List[str]:
        """
        Propagates DOI annotation from contained datasets.
        """
        dois = []
        for _, v in self.datasets.items():
            vdoi = v.doi_journal
            if isinstance(vdoi, str):
                vdoi = [vdoi]
            dois.extend(vdoi)
        return np.unique(dois).tolist()

    @property
    def supplier(self) -> List[str]:
        """
        Propagates supplier annotation from contained datasets.
        """
        supplier = [v.supplier for _, v in self.datasets.items()]
        return np.unique(supplier).tolist()

    def show_summary(self):
        for k, v in self.datasets.items():
            print(k)
            print(f"\t {(v.supplier, v.organism, v.organ, v.assay_sc, v.disease)}")


class DatasetGroupDirectoryOriented(DatasetGroup):

    _cwd: os.PathLike

    def __init__(
            self,
            file_base: str,
            data_path: Union[str, None] = None,
            meta_path: Union[str, None] = None,
            cache_path: Union[str, None] = None,
    ):
        """
        Automatically collects Datasets from python files in directory.

        Uses a pre-built DatasetGroup if this is defined in a group.py file, otherwise, the DatasetGroup is initialised
        here.

        :param file_base:
        :param data_path:
        :param meta_path:
        :param cache_path:
        """
        # Collect all data loaders from files in directory:
        datasets = []
        self._cwd = os.path.dirname(file_base)
        try:
            collection_id = str(self._cwd).split(os.sep)[-1]
            if len(str(self._cwd).split(os.sep)) > 4:
                package_source = str(self._cwd).split(os.sep)[-5]
                if package_source == "sfaira":
                    pass
                elif package_source == "sfaira_extension":
                    package_source = "sfairae"
                else:
                    package_source = "external"
            else:
                package_source = "external"
        except IndexError as e:
            raise IndexError(f"{e} for {self._cwd}")
        loader_pydoc_path_sfaira = "sfaira.data.dataloaders.loaders."
        loader_pydoc_path_sfairae = "sfaira_extension.data.dataloaders.loaders."
        if package_source == "sfaira":
            loader_pydoc_path = loader_pydoc_path_sfaira
        elif package_source == "sfaira_extension":
            loader_pydoc_path = loader_pydoc_path_sfairae
        else:
            external_dir = os.path.dirname(os.path.dirname(file_base))
            if external_dir not in sys.path:
                sys.path.append(external_dir)
            loader_pydoc_path = ""

        # List all files:
        fns = [x for x in os.listdir(self._cwd) if os.path.isfile(os.path.join(self._cwd, x))]
        # Data loader files:
        fns_loader = [x for x in fns
                      if x.split(".")[-1] == "py" and x.split(".")[0] not in ["__init__", "base", "group"]]
        # Ontology class maps .tsvs:
        fns_tsv = [os.path.join(self._cwd, x) for x in fns if x.split(".")[-1] == "tsv"]
        for f in fns_loader:
            datasets_f = []
            file_module = ".".join(f.split(".")[:-1])
            DatasetFound = pydoc.locate(loader_pydoc_path + collection_id + "." + file_module + ".Dataset")
            # Load objects from name space:
            # - load(): Loading function that return anndata instance.
            # - SAMPLE_FNS: File name list for DatasetBaseGroupLoadingManyFiles
            load_func = pydoc.locate(loader_pydoc_path + collection_id + "." + file_module + ".load")
            load_func_annotation = \
                pydoc.locate(loader_pydoc_path + collection_id + "." + file_module + ".LOAD_ANNOTATION")
            # Also check sfaira_extension for additional load_func_annotation:
            if package_source != "sfairae":
                load_func_annotation_sfairae = pydoc.locate(loader_pydoc_path_sfairae + collection_id +
                                                            "." + file_module + ".LOAD_ANNOTATION")
                # LOAD_ANNOTATION is a dictionary so we can use update to extend it.
                if load_func_annotation_sfairae is not None and load_func_annotation is not None:
                    load_func_annotation.update(load_func_annotation_sfairae)
                elif load_func_annotation_sfairae is not None and load_func_annotation is None:
                    load_func_annotation = load_func_annotation_sfairae
            sample_fns = pydoc.locate(loader_pydoc_path + collection_id + "." + file_module +
                                      ".SAMPLE_FNS")
            fn_yaml = os.path.join(self._cwd, file_module + ".yaml")
            fn_yaml = fn_yaml if os.path.exists(fn_yaml) else None
            # Check for sample_fns in yaml:
            if fn_yaml is not None:
                assert os.path.exists(fn_yaml), f"did not find yaml {fn_yaml}"
                yaml_vals = read_yaml(fn=fn_yaml)
                if sample_fns is None and yaml_vals["meta"]["sample_fns"] is not None:
                    sample_fns = yaml_vals["meta"]["sample_fns"]
            if sample_fns is None:
                sample_fns = [None]
            # Here we distinguish between class that are already defined and those that are not.
            # The latter case arises if meta data are defined in YAMLs and _load is given as a function.
            if DatasetFound is None:
                for x in sample_fns:
                    datasets_f.append(
                        DatasetBase(
                            data_path=data_path,
                            meta_path=meta_path,
                            cache_path=cache_path,
                            load_func=load_func,
                            dict_load_func_annotation=load_func_annotation,
                            sample_fn=x,
                            sample_fns=sample_fns if sample_fns != [None] else None,
                            yaml_path=fn_yaml,
                        )
                    )
            else:
                for x in sample_fns:
                    datasets_f.append(
                        DatasetFound(
                            data_path=data_path,
                            meta_path=meta_path,
                            cache_path=cache_path,
                            load_func=load_func,
                            load_func_annotation=load_func_annotation,
                            sample_fn=x,
                            sample_fns=sample_fns if sample_fns != [None] else None,
                            yaml_path=fn_yaml,
                        )
                    )
            # Load ontology label maps:
            for x in datasets_f:
                x.read_ontology_class_maps(fns=fns_tsv)
            datasets.extend(datasets_f)

        keys = [x.id for x in datasets]
        super().__init__(datasets=dict(zip(keys, datasets)), collection_id=collection_id)

    def clean_ontology_class_maps(self):
        """
        Finalises processed class maps of free text labels to ontology classes.

        Checks that the assigned ontology class names appear in the ontology.
        Adds a third column with the corresponding ontology IDs into the file.

        :return:
        """
        for f in os.listdir(self._cwd):
            if os.path.isfile(os.path.join(self._cwd, f)):  # only files
                # Narrow down to data set files:
                if f.split(".")[-1] == "py" and f.split(".")[0] not in ["__init__", "base", "group"]:
                    fn = ".".join(f.split(".")[:-1])
                    # TODO: dectecting both old and new loader versions here, can deprecate in future when all are
                    #  of the new type.
                    # <v1.1 tsvs
                    fns_map_old = [os.path.join(self._cwd, fn + ".tsv")]
                    # >=v1.1 tsvs:
                    fns_map = [os.path.join(self._cwd, fn + "_" + x + ".tsv")
                               for x in self._adata_ids.ontology_constrained]
                    fns_map = fns_map_old + fns_map
                    for fn_map in fns_map:
                        # Get a data set instance from the group to use its methods below.
                        v = self.datasets[self.ids[0]]
                        onto_key = identify_tsv(fn=fn_map, ontology_names=self._adata_ids.ontology_constrained)
                        if os.path.exists(fn_map):
                            onto = v.get_ontology(onto_key)
                            # Access reading and value protection mechanisms from first data set loaded in group.
                            tab = v._read_ontology_class_map(fn=fn_map)
                            # Checks that the assigned ontology class names appear in the ontology.
                            v._value_protection(
                                attr=onto_key,
                                allowed=onto,
                                attempted=[
                                    x for x in np.unique(tab[self._adata_ids.classmap_target_key].values)
                                    if x not in [
                                        self._adata_ids.unknown_metadata_identifier,
                                        self._adata_ids.not_a_cell_celltype_identifier
                                    ]
                                ]
                            )
                            # Adds a third column with the corresponding ontology IDs into the file.
                            tab[self._adata_ids.classmap_target_id_key] = [
                                onto.convert_to_id(x)
                                if (x != self._adata_ids.unknown_metadata_identifier and
                                    x != self._adata_ids.not_a_cell_celltype_identifier)
                                else self._adata_ids.unknown_metadata_identifier
                                for x in tab[self._adata_ids.classmap_target_key].values
                            ]
                            v._write_ontology_class_map(fn=fn_map, tab=tab)


class DatasetSuperGroup:
    """
    Container for multiple DatasetGroup instances.

    Used to manipulate structured dataset collections. Primarly designed for this manipulation, convert to DatasetGroup
    via flatten() for more functionalities.
    """
    _adata: Union[None, anndata.AnnData]
    fn_backed: Union[None, PathLike]
    dataset_groups: Union[list, List[DatasetGroup], List[DatasetSuperGroup]]

    def __init__(self, dataset_groups: Union[None, List[DatasetGroup], List[DatasetSuperGroup]]):
        self._adata = None
        self.fn_backed = None
        self.set_dataset_groups(dataset_groups=dataset_groups)

        self._adata_ids = AdataIdsSfaira()

    def set_dataset_groups(self, dataset_groups: Union[DatasetGroup, DatasetSuperGroup, List[DatasetGroup],
                                                       List[DatasetSuperGroup]]):
        if isinstance(dataset_groups, DatasetGroup) or isinstance(dataset_groups, DatasetSuperGroup):
            dataset_groups = [dataset_groups]
        if len(dataset_groups) > 0:
            if isinstance(dataset_groups[0], DatasetGroup):
                self.dataset_groups = dataset_groups
            elif isinstance(dataset_groups[0], DatasetSuperGroup):
                # Decompose super groups first
                dataset_groups_proc = []
                for x in dataset_groups:
                    dataset_groups_proc.extend(x.dataset_groups)
                self.dataset_groups = dataset_groups_proc
            else:
                assert False
        else:
            self.dataset_groups = []

    def extend_dataset_groups(self, dataset_groups: Union[List[DatasetGroup], List[DatasetSuperGroup]]):
        if isinstance(dataset_groups[0], DatasetGroup):
            self.dataset_groups.extend(dataset_groups)
        elif isinstance(dataset_groups[0], DatasetSuperGroup):
            # Decompose super groups first
            dataset_groups_proc = []
            for x in dataset_groups:
                dataset_groups_proc.extend(x.datasets)
            self.dataset_groups.extend(dataset_groups_proc)
        else:
            assert False

    @property
    def ids(self):
        ids = []
        for x in self.dataset_groups:
            ids.extend(x.ids)
        return ids

    def get_gc(self, genome: str = None):
        g = GenomeContainer(release=genome)
        return g

    def ncells_bydataset(self, annotated_only: bool = False):
        """
        List of list of length of all data sets by data set group.
        :return:
        """
        return [x.ncells_bydataset(annotated_only=annotated_only) for x in self.dataset_groups]

    def ncells_bydataset_flat(self, annotated_only: bool = False):
        """
        Flattened list of length of all data sets.
        :return:
        """
        return [xx for x in self.ncells_bydataset(annotated_only=annotated_only) for xx in x]

    def ncells(self, annotated_only: bool = False):
        return np.sum(self.ncells_bydataset_flat(annotated_only=annotated_only))

    @property
    def datasets(self) -> Dict[str, DatasetBase]:
        """
        Returns DatasetGroup (rather than self = DatasetSuperGroup) containing all listed data sets.

        :return:
        """
        ds = {}
        for x in self.dataset_groups:
            for k, v in x.datasets.items():
                assert k not in ds.keys(), f"{k} was duplicated in super group, remove duplicates before flattening"
                ds[k] = v
        return ds

    def flatten(self) -> DatasetGroup:
        """
        Returns DatasetGroup (rather than self = DatasetSuperGroup) containing all listed data sets.

        :return:
        """
        ds = {}
        for x in self.dataset_groups:
            for k, v in x.datasets.items():
                assert k not in ds.keys(), f"{k} was duplicated in super group, purge duplicates before flattening"
                ds[k] = v
        return DatasetGroup(datasets=ds)

    def download(self, **kwargs):
        for x in self.dataset_groups:
            x.download(**kwargs)

    def load(
            self,
            annotated_only: bool = False,
            load_raw: bool = False,
            allow_caching: bool = True,
            processes: int = 1,
            **kwargs
    ):
        """
        Loads data set homosapiens into anndata object.

        :param annotated_only:
        :param load_raw: See .load().
        :param allow_caching: See .load().
        :param processes: Processes to parallelise loading over. Uses python multiprocessing if > 1, for loop otherwise.
            Note: parallelises loading of each dataset group, but not across groups.
        :return:
        """
        for x in self.dataset_groups:
            x.load(
                annotated_only=annotated_only,
                load_raw=load_raw,
                allow_caching=allow_caching,
                processes=processes,
                **kwargs
            )

    def streamline_features(
            self,
            match_to_release: Union[str, Dict[str, str], None] = None,
            remove_gene_version: bool = True,
            subset_genes_to_type: Union[None, str, List[str]] = None,
            schema: Union[str, None] = None,
    ):
        """
        Subset and sort genes to genes defined in an assembly or genes of a particular type, such as protein coding.

        :param match_to_release: Which genome annotation release to map the feature space to. Note that assemblies from
            ensembl are usually named as Organism.Assembly.Release, this is the Release string. Can be:
                - str: Provide the name of the release.
                - dict: Mapping of organism to name of the release (see str format). Chooses release for each
                    data set based on organism annotation.:param remove_gene_version: Whether to remove the version
                    number after the colon sometimes found in ensembl gene ids.
        :param subset_genes_to_type: Type(s) to subset to. Can be a single type or a list of types or None.
            Types can be:
                - None: All genes in assembly.
                - "protein_coding": All protein coding genes in assembly.
        """
        for x in self.dataset_groups:
            x.streamline_features(
                match_to_release=match_to_release,
                remove_gene_version=remove_gene_version,
                subset_genes_to_type=subset_genes_to_type,
                schema=schema,
            )

    def collapse_counts(self):
        """
        Collapse count matrix along duplicated index.
        """
        for x in self.dataset_groups:
            x.collapse_counts()

    @property
    def adata_ls(self):
        adata_ls = []
        for k, v in self.flatten().datasets.items():
            if v.adata is not None:
                adata_ls.append(v.adata)
        return adata_ls

    @property
    def adata(self):
        adata_ls = self.adata_ls
        if not adata_ls:
            return None
        if len(adata_ls) == 1:
            for i in self.ids:
                if self.datasets[i] is not None:
                    if self.datasets[i].adata is not None:
                        ds_id = i
            adata_concat = adata_ls[0]
            adata_concat.obs[self._adata_ids.dataset] = ds_id
        else:
            # Check that all individual adata objects in linked Dataset instances have identicall streamlined features and metadata
            match_ref_list = []
            rm_gene_ver_list = []
            gene_type_list = []
            for d_id in self.flatten().ids:
                if self.flatten().datasets[d_id].adata is not None:
                    assert self.flatten().datasets[d_id].mapped_features, f"Dataset {d_id} does not seem to have a streamlined " \
                                                                          f"featurespace. To obtain an adata object from this " \
                                                                          f"DatasetGroup, all contained Datasets need to have a " \
                                                                          f"streamlined featurespace. Run .streamline_features()" \
                                                                          f" first."
                    assert self.flatten().datasets[d_id].streamlined_meta, f"Dataset {d_id} does not seem to have streamlined " \
                                                                           f"metadata. To obtain an adata object from this " \
                                                                           f"DatasetGroup, all contained Datasets need to have " \
                                                                           f"streamlined metadata. Run .streamline_metadata() first."
                    match_ref_list.append(self.flatten().datasets[d_id].mapped_features)
                    rm_gene_ver_list.append(self.flatten().datasets[d_id].remove_gene_version)
                    gene_type_list.append(self.flatten().datasets[d_id].subset_gene_type)
            assert len(set(match_ref_list)) == 1, \
                "Not all datasets in this group had their features matched to the same reference (argument " \
                "'match_to_reference' of method .streamline_features()). This is however a prerequisite for creating a " \
                "combined adata object."
            assert len(set(rm_gene_ver_list)) == 1, \
                "Not all datasets in this group have had their gene version removed (argument 'remove_gene_version' of " \
                "method .streamline_features()). This is however a prerequisite for creating a combined adata object."
            assert len(set(gene_type_list)) == 1, \
                "Not all datasets in this group had their featurespace subsetted to the same gene type (argument " \
                "'subset_gene_type' of method .streamline_features()). This is however a prerequisite for creating a " \
                "combined adata object."

            var_original = adata_ls[0].var.copy()
            for a in adata_ls:
                a.var_names_make_unique()
            # TODO: need to keep this? -> yes, still catching errors here (March 2020)
            # Fix for loading bug: sometime concatenating sparse matrices fails the first time but works on second try.
            try:
                adata_concat = adata_ls[0].concatenate(
                    *adata_ls[1:],
                    join="outer",
                    batch_key=self._adata_ids.dataset,
                    batch_categories=[i for i in self.ids if self.flatten().datasets[i].adata is not None],
                    index_unique=None
                )
            except ValueError:
                adata_concat = adata_ls[0].concatenate(
                    *adata_ls[1:],
                    join="outer",
                    batch_key=self._adata_ids.dataset,
                    batch_categories=[i for i in self.ids if self.flatten().datasets[i].adata is not None],
                    index_unique=None
                )
            adata_concat.var = var_original
            adata_concat.uns = merge_uns_from_list(adata_ls)
            adata_concat.uns[self._adata_ids.mapped_features] = match_ref_list[0]

        return adata_concat

    def write_distributed_store(
            self,
            dir_cache: Union[str, os.PathLike],
            store_format: str = "dao",
            dense: bool = False,
            compression_kwargs: dict = {},
            chunks: Union[int, None] = None,
    ):
        """
        Write data set into a format that allows distributed access to data set on disk.

        Stores are useful for distributed access to data sets, in many settings this requires some streamlining of the
        data sets that are accessed. Use .streamline_* before calling this method to streamline the data sets.
        This method writes a separate file for each data set in this object.

        :param dir_cache: Directory to write cache in.
        :param store_format: Disk format for objects in cache. Recommended is "dao".

            - "h5ad": Allows access via backed .h5ad.
                Note on compression: .h5ad supports sparse data with is a good compression that gives fast row-wise
                    access if the files are csr, so further compression potentially not necessary.
            - "dao": Distributed access optimised format, recommended for batched access in optimisation, for example.
        :param dense: Whether to write sparse or dense store, this will be homogenously enforced.
        :param compression_kwargs: Compression key word arguments to give to h5py or zarr
            For store_format=="h5ad", see also anndata.AnnData.write_h5ad:
                - compression,
                - compression_opts.
            For store_format=="dao", see also sfaira.data.write_dao which relays kwargs to
            zarr.hierarchy.create_dataset:
                - compressor
                - overwrite
                - order
                and others.
        :param chunks: Observation axes of chunk size of zarr array, see anndata.AnnData.write_zarr documentation.
            Only relevant for store=="dao". The feature dimension of the chunks is always is the full feature space.
            Uses zarr default chunking across both axes if None.
        """
        for x in self.dataset_groups:
            x.write_distributed_store(dir_cache=dir_cache, store_format=store_format, dense=dense,
                                      compression_kwargs=compression_kwargs, chunks=chunks)

    def streamline_metadata(
            self,
            schema: str = "sfaira",
            clean_obs: bool = True,
            clean_var: bool = True,
            clean_uns: bool = True,
            clean_obs_names: bool = True,
            keep_orginal_obs: bool = False,
            keep_symbol_obs: bool = True,
            keep_id_obs: bool = True,
    ):
        """
        Streamline the adata instance in each group and each data set to output format.
        Output format are saved in ADATA_FIELDS* classes.

        :param schema: Export format.
            - "sfaira"
            - "cellxgene"
        :param clean_obs: Whether to delete non-streamlined fields in .obs, .obsm and .obsp.
        :param clean_var: Whether to delete non-streamlined fields in .var, .varm and .varp.
        :param clean_uns: Whether to delete non-streamlined fields in .uns.
        :param clean_obs_names: Whether to replace obs_names with a string comprised of dataset id and an increasing integer.
        :param clean_obs_names: Whether to replace obs_names with a string comprised of dataset id and an increasing
            integer.
        :param keep_orginal_obs: For ontology-constrained .obs columns, whether to keep a column with original
            annotation.
        :param keep_symbol_obs: For ontology-constrained .obs columns, whether to keep a column with ontology symbol
            annotation.
        :param keep_id_obs: For ontology-constrained .obs columns, whether to keep a column with ontology ID annotation.
        :return:
        """
        for x in self.dataset_groups:
            for xx in x.ids:
                x.datasets[xx].streamline_metadata(
                    schema=schema,
                    clean_obs=clean_obs,
                    clean_var=clean_var,
                    clean_uns=clean_uns,
                    clean_obs_names=clean_obs_names,
                    keep_orginal_obs=keep_orginal_obs,
                    keep_symbol_obs=keep_symbol_obs,
                    keep_id_obs=keep_id_obs,
                )

    def subset(self, key, values):
        """
        Subset list of adata objects based on match to values in key property.

        These keys are properties that are available in lazy model.
        Subsetting happens on .datasets.

        :param key: Property to subset by.
        :param values: Classes to overlap to.
        :return:
        """
        for x in self.dataset_groups:
            x.subset(key=key, values=values)
        self.dataset_groups = [x for x in self.dataset_groups if x.datasets]  # Delete empty DatasetGroups

    def remove_duplicates(
            self,
            supplier_hierarchy: str = "cellxgene,sfaira"
    ):
        """
        Remove duplicate data loaders from super group, e.g. loaders that map to the same DOI.

        Any DOI match is removed (pre-print or journal publication).
        Data sets without DOI are removed, too.
        Loaders are kept in the hierarchy indicated in supplier_hierarchy.
        Requires a super group with homogenous suppliers across DatasetGroups, throws an error otherwise.
        This is given for sfaira maintained libraries but may not be the case if custom assembled DatasetGroups are
        used.

        :param supplier_hierarchy: Hierarchy to resolve duplications by.
            Comma separated string that indicates which data provider takes priority.
            Choose "cellxgene,sfaira" to prioritise use of data sets downloaded from cellxgene.
            Choose "sfaira,cellxgene" to prioritise use of raw data processing pipelines locally.

                - cellxgene: cellxgene downloads
                - sfaira: local raw file processing
        :return:
        """
        # Build a pairing of provider and DOI:
        report_list = []
        idx_tokeep = []
        supplier_hierarchy = supplier_hierarchy.split(",")
        for i, (x, y) in enumerate([(xx.supplier, xx.doi) for xx in self.dataset_groups]):
            if len(x) > 1:
                raise ValueError(f"found more than one supplier for DOI {str(y)}")
            else:
                x = x[0]
                if x not in supplier_hierarchy:
                    raise ValueError(f"could not associate supplier {x} with hierarchy {supplier_hierarchy} in "
                                     f"data set {y}")
                if len(report_list) > 0:
                    matched_idx = np.where([
                        np.any([
                            zz in y
                            for zz in z[1]
                        ])
                        for z in report_list
                    ])[0]
                    assert len(matched_idx) < 1, f"more matches than expected for {(x, y)}"
                else:
                    matched_idx = []
                if len(matched_idx) > 0:
                    # Establish which entry takes priority
                    supplier_old = report_list[matched_idx[0]][0]
                    priority = supplier_hierarchy.index(supplier_old) > supplier_hierarchy.index(x)
                    print(f"removing duplicate data set {y} from supplier: {supplier_old if priority else x}")
                    if priority:
                        idx_tokeep.append(i)
                        del idx_tokeep[matched_idx[0]]
                else:
                    report_list.append([x, y])
                    idx_tokeep.append(i)
        self.dataset_groups = [self.dataset_groups[i] for i in idx_tokeep]

    def subset_cells(self, key, values: Union[str, List[str]]):
        """
        Subset list of adata objects based on cell-wise properties.

        These keys are properties that are not available in lazy model and require loading first because the
        subsetting works on the cell-level: .adata are maintained but reduced to matches.

        :param key: Property to subset by. Options:

            - "assay_sc" points to self.assay_sc_obs_key
            - "assay_differentiation" points to self.assay_differentiation_obs_key
            - "assay_type_differentiation" points to self.assay_type_differentiation_obs_key
            - "cell_line" points to self.cell_line
            - "cell_type" points to self.cell_type_obs_key
            - "developmental_stage" points to self.developmental_stage_obs_key
            - "ethnicity" points to self.ethnicity_obs_key
            - "organ" points to self.organ_obs_key
            - "organism" points to self.organism_obs_key
            - "sample_source" points to self.sample_source_obs_key
            - "sex" points to self.sex_obs_key
            - "state_exact" points to self.state_exact_obs_key
        :param values: Classes to overlap to.
        :return:
        """
        for i in range(len(self.dataset_groups)):
            self.dataset_groups[i].subset_cells(key=key, values=values)

    def project_celltypes_to_ontology(self, adata_fields: Union[AdataIds, None] = None, copy=False):
        """
        Project free text cell type names to ontology based on mapping table.
        :return:
        """
        for _, v in self.dataset_groups:
            v.project_free_to_ontology(adata_fields=adata_fields, copy=copy)

    def write_config(self, fn: Union[str, os.PathLike]):
        """
        Writes a config file that describes the current data sub-setting.

        This config file can be loaded later to recreate a sub-setting.

        :param fn: Output file.
        """
        pd.DataFrame({"id": np.sort(self.ids)}).to_csv(fn, index=False, sep="\t")

    def load_config(self, fn: Union[str, os.PathLike]):
        """
        Load a config file and recreates a data sub-setting.

        :param fn: Output file.
        """
        tab = pd.read_csv(fn, header=0, index_col=None, sep="\t")
        ids_keep = tab["id"].values
        self.subset(key="id", values=ids_keep)

    @property
    def additional_annotation_key(self) -> List[Dict[str, Union[None, str]]]:
        """"
        Return list (by data set group) of dictionaries of additional_annotation_key for each data set with ids as keys.
        """
        return [
            dict([
                (k, x.datasets[k].additional_annotation_key)
                for k, v in x.datasets.items()
            ]) for x in self.dataset_groups
        ]

    @additional_annotation_key.setter
    def additional_annotation_key(self, x: Dict[str, Union[None, str]]):
        """
        Allows setting of additional_annotation_key in a subset of datasets identifed by keys in x.

        The input is not structured by DatasetGroups but only by ID, all groups are checked for matching IDs.

        :param x: Dictionary with data set ids in keys and new _additional_annotation_key values to be setted in values.
            Note that you can either add  or change secondary annotation by setting a value to a string or remove it
            by setting a value to None.
        :return:
        """
        for k, v in x.items():
            counter = 0
            for x in self.dataset_groups:
                if k in x.ids:
                    x.datasets[k].additional_annotation_key = v
                    counter += 1
            if counter == 0:
                warnings.warn(f"did not data set matching ID {k}")
            elif counter > 1:
                warnings.warn(f"found more than one ({counter}) data set matching ID {k}")

    def show_summary(self):
        for k, v in self.datasets.items():
            print(k)
            print(f"\t {(v.supplier, v.organism, v.organ, v.assay_sc, v.disease)}")



import numpy as np
import pytest
from typing import Tuple, Union

from sfaira.versions.genomes import GenomeContainer, translate_id_to_symbols, translate_symbols_to_id

RELEASE = "105"
ORGANISM = "Mus musculus"

"""
GenomeContainer.
"""


def test_gc_init():
    """
    Tests different modes of initialisation for fatal errors.
    """
    gc = GenomeContainer(release=RELEASE, organism=ORGANISM)
    assert gc.organism == "Mus musculus"


@pytest.mark.parametrize("subset", [
    ({"biotype": "protein_coding"}, 21936),
    ({"biotype": "lincRNA"}, 5629),
    ({"biotype": "protein_coding,lincRNA"}, 21936 + 5629),
    ({"symbols": "Gnai3,Pbsn,Cdc45"}, 3),
    ({"ensg": "ENSMUSG00000000003,ENSMUSG00000000028"}, 2)
])
def test_gc_subsetting(subset: Tuple[dict, int]):
    """
    Tests if genome container is subsetted correctly.
    """
    gc = GenomeContainer(release="102", organism=ORGANISM)
    gc.set(**subset[0])
    assert gc.n_var == subset[1]
    assert len(gc.ensembl) == subset[1]
    assert len(gc.symbols) == subset[1]
    assert len(gc.biotype) == subset[1]
    if list(subset[0].keys())[0] == "protein_coding":
        assert np.all(gc.biotype == "protein_coding")


@pytest.mark.parametrize("organism", ["Caenorhabditis elegans",
                                      "Callithrix jacchus",
                                      "Sus scrofa"])
def test_gc_nonstandard_organism(organism: str):
    """
    Tests if genome container works for selected non-standard organism.
    """
    gc = GenomeContainer(release=RELEASE, organism=organism)
    assert gc.n_var > 10


"""
Utils.
"""


@pytest.mark.parametrize("genes", [
    ("Adora3", "ENSMUSG00000000562"),  # single string
    (["Adora3", "Timp1"], ["ENSMUSG00000000562", "ENSMUSG00000001131"]),  # list of strings
    (["ADORA3", "timp1"], ["EnsmusG00000000562", "ENSMUSG00000001131"]),  # list of strings with weird capitalization
])
def test_translate_id_to_symbols(genes):
    """
    Tests translate_id_to_symbols and translate_symbols_to_id for translation errors.
    """
    x, y = genes
    y_hat = translate_symbols_to_id(x=x, release=RELEASE, organism=ORGANISM)
    # Correct target spelling of y:
    y = [z.upper() for z in y] if isinstance(y, list) else y.upper()
    assert np.all(y_hat == y)
    y, x = genes
    y_hat = translate_id_to_symbols(x=x, release=RELEASE, organism=ORGANISM)
    # Correct target spelling of y:
    y = [z[0].upper() + z[1:].lower() for z in y] if isinstance(y, list) else y[0].upper() + y[1:].lower()
    assert np.all(y_hat == y)
